/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "CloseConfirmation.h"
#include <stdexcept>

class CloseConfirmationDialog::Columns: public Gtk::TreeModelColumnRecord
{
public:
  Columns()
  {
    add(save);
    add(title);
    add(document);
  }

  Gtk::TreeModelColumn<bool> save;
  Gtk::TreeModelColumn<Glib::ustring> title;
  Gtk::TreeModelColumn<Document*> document;
};

CloseConfirmationDialog::iterator::iterator(const Gtk::TreeIter& init, Columns& columns):
  m_columns(columns), m_internal(init)
{
}

CloseConfirmationDialog::iterator::iterator(const iterator& other):
  m_columns(other.m_columns), m_internal(other.m_internal)
{
}

CloseConfirmationDialog::iterator& CloseConfirmationDialog::iterator::operator++()
{
  ++ m_internal;
  return *this;
}

CloseConfirmationDialog::iterator CloseConfirmationDialog::iterator::operator++(int)
{
  iterator temp(*this);
  ++ m_internal;
  return temp;
}

bool CloseConfirmationDialog::iterator::operator==(const iterator& other)
{
  return m_internal == other.m_internal;
}

bool CloseConfirmationDialog::iterator::operator!=(const iterator& other)
{
  return m_internal != other.m_internal;
}

Document& CloseConfirmationDialog::iterator::get_document()
{
  return *(*m_internal)[m_columns.document];
}

bool CloseConfirmationDialog::iterator::get_save()
{
  return (*m_internal)[m_columns.save];
}

struct CloseConfirmationDialog::Private
{
  Private(Gtk::Window& parent_):
    parent(parent_), list(Gtk::ListStore::create(columns)),
    filter(Gtk::TreeModelFilter::create(list)),
    view(filter)
  {
    filter->set_visible_func(sigc::mem_fun(*this, &Private::on_visible));

    view.append_column_editable("Save changes?", columns.save);
    view.append_column("Document title", columns.title);
    view.set_headers_visible(false);
    view.show();

    scrolled_window.add(view);
    scrolled_window.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
    scrolled_window.set_shadow_type(Gtk::SHADOW_IN);
    scrolled_window.show();
  }

  bool on_visible(const Gtk::TreeIter& iter)
  {
    Document* doc = (*iter)[columns.document];
    if(doc == NULL) return false; // Not yet initialized
    return doc->get_text_buffer()->get_modified();
  }

  Gtk::Window& parent;

  Columns columns;
  Glib::RefPtr<Gtk::ListStore> list;
  Glib::RefPtr<Gtk::TreeModelFilter> filter;

  Gtk::ScrolledWindow scrolled_window;
  Gtk::TreeView view;
};

CloseConfirmationDialog::CloseConfirmationDialog(Gtk::Window& parent):
  Gtk::MessageDialog(parent, "", false, Gtk::MESSAGE_WARNING, Gtk::BUTTONS_NONE, true),
  m_priv(new Private(parent))
{
  set_title("Save changed documents");
  Gtk::Button* button = add_button(Gtk::Stock::CLOSE, Gtk::RESPONSE_CLOSE);
  button->set_label("Close _without Saving");

  add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
  button = add_button(Gtk::Stock::SAVE, Gtk::RESPONSE_ACCEPT);
  button->grab_focus();

  get_vbox()->pack_start(m_priv->scrolled_window);
}

CloseConfirmationDialog::~CloseConfirmationDialog()
{
  // Make the compiler happy
}

void CloseConfirmationDialog::add_document(Document& doc)
{
  // TODO: Watch for document removal, title change or modified flag change
  Gtk::TreeIter iter = m_priv->list->append();
  (*iter)[m_priv->columns.save] = doc.get_text_buffer()->get_modified();
  (*iter)[m_priv->columns.title] = doc.get_title();
  (*iter)[m_priv->columns.document] = &doc;
}

void CloseConfirmationDialog::clear()
{
  m_priv->list->clear();
}

CloseConfirmationDialog::Action CloseConfirmationDialog::launch()
{
  // Count non-modified documents
  unsigned int count = 0;
  Document* last_modified_doc = NULL;
  gchar* text;

  Gtk::TreeNodeChildren children = m_priv->list->children();
  for(Gtk::TreeIter iter = children.begin(); iter != children.end(); ++ iter)
  {
    Document* doc = (*iter)[m_priv->columns.document];
    if(doc->get_text_buffer()->get_modified())
    {
      last_modified_doc = doc;
      ++ count;
    }
  }

  if(count == 0) return CLOSE_WITHOUT_SAVE;

  if(count == 1)
  {
    text = g_strdup_printf("Save the changes to document \"%s\" before closing?", last_modified_doc->get_title().c_str());
    set_message(text);
    g_free(text);
  }
  else
  {
    text = g_strdup_printf("Save changes to %d unsaved documents before closing?", count);
    set_message(text);
    g_free(text);
  }

  set_secondary_text("If you don't save, changes will be permanently lost.");

  if(count > 1)
    m_priv->scrolled_window.show();
  else
    m_priv->scrolled_window.hide();

  int response = run();
  hide();

  switch(response)
  {
  case Gtk::RESPONSE_CANCEL:
    return CANCEL;
  case Gtk::RESPONSE_CLOSE:
    return CLOSE_WITHOUT_SAVE;
  case Gtk::RESPONSE_ACCEPT:
    return SAVE_SELECTED;
  default:
    throw std::logic_error("CloseConfirmationDialog::launch: Invalid response");
  }
}

CloseConfirmationDialog::iterator CloseConfirmationDialog::begin()
{
  Gtk::TreeNodeChildren children = m_priv->list->children();
  return iterator(children.begin(), m_priv->columns);
}

CloseConfirmationDialog::iterator CloseConfirmationDialog::end()
{
  Gtk::TreeNodeChildren children = m_priv->list->children();
  return iterator(children.end(), m_priv->columns);
}
