/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-scriptview.h>
#include <c4scripter/c4s-textutil.h>

#include "Document.h"

const gchar** keywords_manip(const Document::KeywordList& keywords)
{
  // TODO: Free this. Do it! Wrap KeywordList next!
  const gchar** keys = (const gchar**)g_malloc(sizeof(gchar*) * (keywords.size() + 1));
  for(size_t i = 0; i < keywords.size(); ++ i)
    keys[i] = g_strdup(keywords[i].c_str());
  keys[keywords.size()] = NULL;
  return keys;
}

Document::Document(const KeywordList& keywords, C4sPreferences* preferences, C4sEngineFunctionList* engine_function_list):
  Gtk::ScrolledWindow(), m_preferences(preferences),
  m_title("Untitled"), m_encoding("LATIN1"), m_line_ending(DOS),
  m_textbuffer(GTK_TEXT_BUFFER(gtk_source_buffer_new(NULL))),
  m_highlighter(c4s_highlighter_new(m_textbuffer, keywords_manip(keywords), engine_function_list, preferences)),
  m_function_list(c4s_custom_function_list_new(m_textbuffer, m_highlighter)),
  m_param_info(c4s_param_info_new(m_textbuffer, m_highlighter, engine_function_list, m_function_list)),
  m_sourceview(c4s_script_view_new(GTK_SOURCE_BUFFER(m_textbuffer), m_param_info, m_preferences))
{
  gtk_source_view_set_show_line_numbers(GTK_SOURCE_VIEW(m_sourceview), TRUE);
  gtk_source_view_set_show_margin(GTK_SOURCE_VIEW(m_sourceview), TRUE);
  gtk_container_add(GTK_CONTAINER(gobj()), m_sourceview);
  set_shadow_type(Gtk::SHADOW_IN);

  init_preferences();
}

Document::Document(const std::string& file, const KeywordList& keywords, C4sPreferences* preferences, C4sEngineFunctionList* engine_function_list):
  Gtk::ScrolledWindow(), m_preferences(preferences),
  m_title(Glib::path_get_basename(file)), m_path(file), m_encoding("UTF-8"), m_line_ending(UNKNOWN),
  m_textbuffer(GTK_TEXT_BUFFER(gtk_source_buffer_new(NULL))),
  m_highlighter(c4s_highlighter_new(m_textbuffer, keywords_manip(keywords), engine_function_list, preferences)),
  m_function_list(c4s_custom_function_list_new(m_textbuffer, m_highlighter)),
  m_param_info(c4s_param_info_new(m_textbuffer, m_highlighter, engine_function_list, m_function_list)),
  m_sourceview(c4s_script_view_new(GTK_SOURCE_BUFFER(m_textbuffer), m_param_info, preferences))
{
  gtk_source_view_set_show_line_numbers(GTK_SOURCE_VIEW(m_sourceview), TRUE);
  gtk_source_view_set_show_margin(GTK_SOURCE_VIEW(m_sourceview), TRUE);
  gtk_source_buffer_begin_not_undoable_action(GTK_SOURCE_BUFFER(gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_sourceview))));

  gtk_container_add(GTK_CONTAINER(gobj()), m_sourceview);

  std::string content = Glib::file_get_contents(file);

  // Guess line ending from first line ending in file
  /*std::string::size_type lepos = content.find_first_of("\r\n");
  if(lepos != std::string::npos)
  {
    if(content[lepos] == '\n')
      m_line_ending = UNIX;
    else if(lepos < content.length() - 1 && content[lepos + 1] == '\n')
      m_line_ending = DOS;
    else
      m_line_ending = MAC;
  }*/

  if(g_utf8_validate(content.c_str(), content.length(), NULL) == FALSE)
  {
    // Text is not UTF-8, try LATIN1.
    // Conversion might throw Glib::ConvertError, should be catched above in the callstack
    content = Glib::convert(content, "UTF-8", "LATIN1");
    m_encoding = "LATIN1";
  }
  else
  {
    // Valid UTF-8, but may also be valid LATIN1
    if(c4s_text_is_ascii(content.c_str()) )
      m_encoding = "LATIN1";
  }

  std::string::size_type prev = 0, pos;
  GtkTextIter iter; gtk_text_buffer_get_start_iter(gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_sourceview)), &iter);

  while((pos = content.find_first_of("\r\n", prev)) != std::string::npos)
  {
    std::string sub = content.substr(prev, pos - prev);

    gtk_text_buffer_insert(
      gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_sourceview)),
      &iter,
      sub.c_str(),
      sub.length()
    );

    gtk_text_buffer_insert(
      gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_sourceview)),
      &iter,
      "\n",
      1
    );

    if(content[pos] == '\r')
    {
      ++ pos;
      if(pos < content.length() && content[pos] == '\n')
      {
        ++ pos;
	m_line_ending = DOS;
      }
      else
      {
        m_line_ending = MAC;
      }
    }
    else
    {
      g_assert(content[pos] == '\n');
      ++ pos;
      m_line_ending = UNIX;
    }

    prev = pos;
  }

  std::string sub = content.substr(prev);
  gtk_text_buffer_insert(
    gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_sourceview)),
    &iter,
    sub.c_str(),
    sub.length()
  );

  // Line ending may still be unknown if the document contains only one line
  // without line ending. Default to DOS in this case.
  if(m_line_ending == UNKNOWN) m_line_ending = DOS;

  gtk_text_buffer_set_modified(gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_sourceview)), FALSE);
  gtk_source_buffer_end_not_undoable_action(GTK_SOURCE_BUFFER(gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_sourceview))));

  set_shadow_type(Gtk::SHADOW_IN);
  init_preferences();
}

Document::~Document()
{
  g_object_unref(m_highlighter);
  g_object_unref(m_function_list);
}

void Document::save(const std::string& path)
{
  GError* error = NULL;
  gchar* text;
  GtkTextBuffer* buffer;
  GtkTextIter prev, pos;
	
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_sourceview));
	
  gtk_text_buffer_get_start_iter(buffer, &prev);
  pos = prev;

  std::string content;

  do
  {
    gtk_text_iter_forward_to_line_end(&pos);

    text = gtk_text_buffer_get_slice(buffer, &prev, &pos, FALSE);
    content += text;
    g_free(text);

    switch(m_line_ending)
    {
    case DOS:
      content += "\r\n";
      break;
    case UNIX:
      content += "\n";
      break;
    case MAC:
      content += "\r";
      break;
    }
  } while(gtk_text_iter_forward_line(&prev) == TRUE);
  
  if(m_encoding != "UTF-8")
  {
    try
    {
      content = Glib::convert(content, m_encoding, "UTF-8");
    }
    catch(Glib::ConvertError& e)
    {
      // Could not convert to origignal encoding, keep UTF-8
      m_encoding = "UTF-8";
    }
  }

  g_file_set_contents(path.c_str(), content.c_str(), -1, &error);

  if(error != NULL)
  {
    Glib::ustring message = error->message;
    int code = error->code;
    g_error_free(error);

    throw Glib::FileError(static_cast<Glib::FileError::Code>(code), message);
  }

  m_title = Glib::path_get_basename(path);
  m_path = path;

  m_signal_title_changed.emit();
  m_signal_path_changed.emit();

  gtk_text_buffer_set_modified(buffer, FALSE);
}

const std::string& Document::get_path() const
{
	return m_path;
}

void Document::jump_to_function(const C4sCustomFunction* function)
{
  GtkTextIter begin;
  GtkTextIter end;

  c4s_custom_function_get_begin(function, &begin);
  c4s_custom_function_get_end(function, &end);

  Gtk::TextIter begin_cpp(&begin);

  get_text_buffer()->select_range(Gtk::TextIter(&begin), Gtk::TextIter(&end));
  get_text_view()->scroll_to(begin_cpp, 0.1, 0.1, 0.1);
  get_text_view()->grab_focus();
}

void Document::init_preferences()
{
  // Called from both ctors
  g_signal_connect(
    G_OBJECT(m_preferences),
    "notify",
    G_CALLBACK(on_notify),
    this
  );

/*  m_preferences.color_text.signal_changed().connect(sigc::mem_fun(*this, &Document::on_color_changed));
  m_preferences.tab_width.signal_changed().connect(sigc::mem_fun(*this, &Document::on_tab_width_changed));
  m_preferences.auto_indentation.signal_changed().connect(sigc::mem_fun(*this, &Document::on_auto_indentation_changed));
  m_preferences.highlight_bracket.signal_changed().connect(sigc::mem_fun(*this, &Document::on_highlight_bracket_changed));
  m_preferences.font.signal_changed().connect(sigc::mem_fun(*this, &Document::on_font_changed));*/

  Gtk::TextView* view = get_text_view();
  view->modify_text(Gtk::STATE_NORMAL, Gdk::Color((GdkColor*)c4s_preferences_get_color_text(m_preferences), true));
  view->modify_text(Gtk::STATE_PRELIGHT, Gdk::Color((GdkColor*)c4s_preferences_get_color_text(m_preferences), true));
  view->modify_text(Gtk::STATE_ACTIVE, Gdk::Color((GdkColor*)c4s_preferences_get_color_text(m_preferences), true));

  gtk_source_view_set_tabs_width(GTK_SOURCE_VIEW(m_sourceview), c4s_preferences_get_tab_width(m_preferences));
  gtk_source_view_set_auto_indent(GTK_SOURCE_VIEW(m_sourceview), c4s_preferences_get_auto_indentation(m_preferences));
  gtk_source_buffer_set_check_brackets(GTK_SOURCE_BUFFER(gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_sourceview))), c4s_preferences_get_highlight_brackets(m_preferences));

  PangoFontDescription* font = pango_font_description_copy(c4s_preferences_get_font(m_preferences));
  gtk_widget_modify_font(m_sourceview, font);
  pango_font_description_free(font);
}

void Document::on_notify(GObject* object, GParamSpec* pspec, gpointer data)
{
  if(strcmp(pspec->name, "color-text") == 0)
    static_cast<Document*>(data)->on_color_changed();
  else if(strcmp(pspec->name, "tab-width") == 0)
    static_cast<Document*>(data)->on_tab_width_changed();
  else if(strcmp(pspec->name, "auto-indentation") == 0)
    static_cast<Document*>(data)->on_auto_indentation_changed();
  else if(strcmp(pspec->name, "highlight-brackets") == 0)
    static_cast<Document*>(data)->on_highlight_bracket_changed();
  else if(strcmp(pspec->name, "font") == 0)
    static_cast<Document*>(data)->on_font_changed();
}

void Document::on_color_changed()
{
  Gtk::TextView* view = get_text_view();
  view->modify_text(Gtk::STATE_NORMAL, Gdk::Color((GdkColor*)c4s_preferences_get_color_text(m_preferences), true));
  view->modify_text(Gtk::STATE_PRELIGHT, Gdk::Color((GdkColor*)c4s_preferences_get_color_text(m_preferences), true));
  view->modify_text(Gtk::STATE_ACTIVE, Gdk::Color((GdkColor*)c4s_preferences_get_color_text(m_preferences), true));
}

void Document::on_tab_width_changed()
{
  gtk_source_view_set_tabs_width(GTK_SOURCE_VIEW(m_sourceview), c4s_preferences_get_tab_width(m_preferences));
}

void Document::on_auto_indentation_changed()
{
  gtk_source_view_set_auto_indent(GTK_SOURCE_VIEW(m_sourceview), c4s_preferences_get_auto_indentation(m_preferences));
}

void Document::on_highlight_bracket_changed()
{
  gtk_source_buffer_set_check_brackets(GTK_SOURCE_BUFFER(gtk_text_view_get_buffer(GTK_TEXT_VIEW(m_sourceview))), c4s_preferences_get_highlight_brackets(m_preferences));
}

void Document::on_font_changed()
{
  PangoFontDescription* font = pango_font_description_copy(c4s_preferences_get_font(m_preferences));
  gtk_widget_modify_font(m_sourceview, font);
  pango_font_description_free(font);
}
