/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdexcept>
#include <fstream>

#ifdef _WIN32
# include <windows.h>
#endif

#include "FileHolder.h"

namespace
{
  class TabLabel: public Gtk::HBox
  {
  public:
    typedef Glib::SignalProxy0<void> SignalClicked;
    TabLabel(Document& document);

    SignalClicked signal_close() { return m_button.signal_clicked(); }
  private:
    void on_title_changed();
    void on_modified_changed();

    void update_label();

    Document& m_document;

    Gtk::Label m_label;
    Gtk::Button m_button;
    Gtk::Image m_image;
  };

  TabLabel::TabLabel(Document& document):
    m_document(document),
    m_image(Gtk::Stock::CLOSE, Gtk::ICON_SIZE_MENU)
  {
    // Set correct label
    update_label();

    // Lookup icon size
    int width, height;
    Gtk::IconSize::lookup(Gtk::ICON_SIZE_MENU, width, height);

    // Resize button to image's size
    m_button.set_size_request(width + 4, height + 4);

    document.get_text_buffer()->signal_modified_changed().connect(sigc::mem_fun(*this, &TabLabel::on_modified_changed));
    document.signal_title_changed().connect(sigc::mem_fun(*this, &TabLabel::on_title_changed));

    m_button.add(m_image);
    m_button.set_relief(Gtk::RELIEF_NONE);

    pack_start(m_label, Gtk::PACK_SHRINK);
    pack_end(m_button, Gtk::PACK_SHRINK);
  }

  void TabLabel::on_title_changed()
  {
    update_label();
  }

  void TabLabel::on_modified_changed()
  {
    update_label();
  }

  void TabLabel::update_label()
  {
    bool modified = m_document.get_text_buffer()->get_modified();
    m_label.set_text((modified ? "*" : "") + m_document.get_title());
  }

  bool load_keywords_from_file(Document::KeywordList& keywords, std::ifstream& stream)
  {
    std::string line;
    while(std::getline(stream, line) )
    {
      if(line.empty() || line[0] == ';') continue;
      keywords.push_back(line);
    }

    return true;
  }

  bool load_keywords(Document::KeywordList& keywords)
  {
    // Use some glib magic to get the folder were the keyword file
    // should be located.
    for(const gchar* const* dirs = g_get_system_data_dirs(); *dirs != NULL; ++ dirs)
    {
      std::string dir = Glib::build_filename(*dirs, "c4scripter");
      std::ifstream stream(Glib::build_filename(dir, "keywords.txt").c_str());
      if(!stream) continue;

      return load_keywords_from_file(keywords, stream);
    }

#ifdef _WIN32
    // Also look at the exe path on Win32
    TCHAR path[MAX_PATH];
    if(!GetModuleFileName(NULL, path, MAX_PATH))
      return false;

    std::string filename = Glib::build_filename(Glib::path_get_dirname(path), "keywords.txt");
    std::ifstream stream(filename.c_str());
    if(!stream) return false;
    return load_keywords_from_file(keywords, stream);
#else
    return false;
#endif
  }
}

FileHolder::FileHolder(C4sPreferences* preferences, Gtk::Window& parent):
  Gtk::Notebook(),
  m_preferences(preferences),
  m_parent(parent),
  m_functions(c4s_engine_function_list_new(parent.gobj(), preferences))
{
  set_scrollable(true);

  if(!load_keywords(m_keywords) )
  {
    Gtk::MessageDialog dialog(parent, "Failed to load keywords", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
    dialog.set_title("Could not load keyword file");
    dialog.set_secondary_text("C4Scripter could not locate the keywords.txt file. Probably C4Scripter has not been installed properly. Make sure the keywords.txt is at a location where C4Scripter will find it.\n\nAs a result keyword highlighting will be disabled.");
    dialog.run();
  }
}

FileHolder::~FileHolder()
{
  g_object_unref(G_OBJECT(m_functions));
}

void FileHolder::on_switch_page(GtkNotebookPage* page, guint page_num)
{
  Gtk::Notebook::on_switch_page(page, page_num);

  if(page_num == static_cast<guint>(-1))
  {
    // Is done below in remove_document
    //m_signal_document_changed.emit(NULL);
  }
  else
  {
    DocumentHolder& holder = *static_cast<DocumentHolder*>(get_nth_page(page_num));
    m_signal_document_changed.emit(&holder);
  }
}

DocumentHolder& FileHolder::create_document()
{
  return create_document(*Gtk::manage(new DocumentHolder(m_keywords, m_functions, m_preferences)), "Untitled");
}

DocumentHolder& FileHolder::create_document(const std::string& path)
{
  return create_document(*Gtk::manage(new DocumentHolder(path, m_keywords, m_functions, m_preferences)), path);
}

DocumentHolder& FileHolder::create_document(DocumentHolder& doc_holder, const std::string& title)
{
  TabLabel* label = Gtk::manage(new TabLabel(doc_holder.get_document()));
  label->show_all();

  // TODO: Should just be a close request to allow close confirmation
  label->signal_close().connect(sigc::bind(sigc::mem_fun(*this, &FileHolder::on_document_close), sigc::ref(doc_holder)));

  append_page(doc_holder, *label);

  doc_holder.show_all();

  m_signal_document_added.emit(doc_holder);

  set_current_page(get_n_pages() - 1);
  doc_holder.get_document().get_text_buffer()->place_cursor(doc_holder.get_document().get_text_buffer()->begin());
  doc_holder.get_document().get_text_view()->grab_focus();

  return doc_holder;
}

void FileHolder::remove_document(DocumentHolder& doc_holder)
{
  m_signal_document_removed.emit(doc_holder);
  remove_page(doc_holder);

  // Corresponding switch_page signal is not triggered, but we have to inform
  // others when the last page has been removed
  if(get_n_pages() == 0)
    m_signal_document_changed(NULL);
}

DocumentHolder* FileHolder::get_current_document()
{
  Gtk::Widget* widget = get_nth_page(get_current_page() );
  return static_cast<DocumentHolder*>(widget);
}

void FileHolder::on_document_close(DocumentHolder& doc_holder)
{
  // Called from TabLabel
  m_signal_document_close.emit(doc_holder);
}
