/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "GotoDialog.h"

GotoDialog::GotoDialog(Gtk::Window& parent, FileHolder& file_holder):
  Gtk::Dialog("Go to line or function", parent),
  m_file_holder(file_holder),
  m_line_button(m_group, "Go to line:"),
  m_function_button(m_group, "Go to function:")
{
  set_type_hint(Gdk::WINDOW_TYPE_HINT_DIALOG);
  set_transient_for(parent);
  set_position(Gtk::WIN_POS_CENTER_ON_PARENT);

  m_renderer_access.property_ypad() = 0;
  m_renderer_name.property_ypad() = 0;

  m_function_box.pack_start(m_renderer_access, false);
  m_function_box.pack_start(m_renderer_name, true);

  m_line_entry.set_activates_default(true);
  m_line_entry.set_digits(0);
  m_line_entry.set_increments(1, 10);

  //m_function_box.set_focus_on_click(true);
  m_line_entry.signal_grab_focus().connect(sigc::mem_fun(*this, &GotoDialog::on_line_focus));
  m_function_box.signal_changed().connect(sigc::mem_fun(*this, &GotoDialog::on_function_focus));

  m_table.attach(m_line_button, 0, 1, 0, 1, Gtk::SHRINK | Gtk::FILL, Gtk::SHRINK);
  m_table.attach(m_function_button, 0, 1, 1, 2, Gtk::SHRINK | Gtk::FILL, Gtk::SHRINK);
  m_table.attach(m_line_entry, 1, 2, 0, 1, Gtk::EXPAND | Gtk::FILL, Gtk::SHRINK);
  m_table.attach(m_function_box, 1, 2, 1, 2, Gtk::EXPAND | Gtk::FILL, Gtk::SHRINK);
  m_table.set_spacings(10);

  get_vbox()->pack_start(m_table, Gtk::PACK_SHRINK);

  set_border_width(10);

  set_resizable(false);
  show_all_children();

  add_button(Gtk::Stock::CLOSE, Gtk::RESPONSE_CLOSE);
  add_button(Gtk::Stock::JUMP_TO, Gtk::RESPONSE_ACCEPT);

  set_default_response(Gtk::RESPONSE_ACCEPT);
}

void GotoDialog::on_response(int response_id)
{
  switch(response_id)
  {
  case Gtk::RESPONSE_CLOSE:
    hide();
    break;
  case Gtk::RESPONSE_ACCEPT:
    if(m_line_button.get_active())
    {
      DocumentHolder* holder = m_file_holder.get_current_document();
      Gtk::TextIter iter = holder->get_document().get_text_buffer()->get_iter_at_line(m_line_entry.get_value_as_int() - 1);
      Gtk::TextIter end = iter; end.forward_line();

      holder->get_document().get_text_buffer()->select_range(iter, end);
      holder->get_document().get_text_view()->scroll_to(iter, 0.1);
    }

    if(m_function_button.get_active() && m_function_box.get_active_row_number() != -1)
    {
      DocumentHolder* holder = m_file_holder.get_current_document();

      Gtk::TreeIter iter = m_function_box.get_active();

      C4sCustomFunction* func;
      gtk_tree_model_get(
        m_function_box.get_model()->gobj(),
	iter.gobj(),
	C4S_FUNCTION_BAR_COLUMN_FUNCTION,
	&func,
	-1
      );
      
      holder->get_document().jump_to_function(func);
    }

    break;
  default:
    Gtk::Dialog::on_response(response_id);
    break;
  }
}

void GotoDialog::on_show()
{
  Gtk::Dialog::on_show();

  DocumentHolder* holder = m_file_holder.get_current_document();

  m_conn_document_changed = m_file_holder.signal_document_changed().connect(sigc::mem_fun(*this, &GotoDialog::on_document_changed));
  on_document_changed(holder);

  if(holder != NULL)
    m_line_entry.set_value(holder->get_document().get_text_buffer()->get_insert()->get_iter().get_line());
}

void GotoDialog::on_hide()
{
  m_conn_document_changed.disconnect();
  m_conn_line_changed.disconnect();

  Gtk::Dialog::on_hide();
}

void GotoDialog::on_line_focus()
{
  m_line_button.set_active(true);
}

void GotoDialog::on_function_focus()
{
  m_function_button.set_active(true);
}

void GotoDialog::on_document_changed(DocumentHolder* holder)
{
  if(holder == NULL)
  {
    m_table.set_sensitive(false);
    set_response_sensitive(Gtk::RESPONSE_ACCEPT, false);
  }
  else
  {
    m_table.set_sensitive(true);
    set_response_sensitive(Gtk::RESPONSE_ACCEPT, true);

    m_conn_line_changed = holder->get_document().get_text_buffer()->signal_changed().connect(sigc::bind(sigc::mem_fun(*this, &GotoDialog::on_line_changed), holder));

    on_line_changed(holder);

    m_function_box.clear_attributes(m_renderer_access);
    m_function_box.clear_attributes(m_renderer_name);

    GtkListStore* list = c4s_function_bar_get_list_store(holder->get_function_bar());
    m_function_box.set_model(Glib::wrap(list, false));

    gtk_cell_layout_add_attribute(
      GTK_CELL_LAYOUT(m_function_box.gobj()),
      GTK_CELL_RENDERER(m_renderer_access.gobj()),
      "pixbuf",
      C4S_FUNCTION_BAR_COLUMN_ACCESS
    );

    gtk_cell_layout_add_attribute(
      GTK_CELL_LAYOUT(m_function_box.gobj()),
      GTK_CELL_RENDERER(m_renderer_name.gobj()),
      "text",
      C4S_FUNCTION_BAR_COLUMN_NAME
    );
  }
}

void GotoDialog::on_line_changed(DocumentHolder* holder)
{
  m_line_entry.set_range(1, holder->get_document().get_text_buffer()->get_line_count());
}
