/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ParamTip.h"
#include <stdexcept>

#ifdef _WIN32
# include <windows.h>
#endif

namespace {
  GdkColor TIP_BORDER_GDK = { 0, 0x0000, 0x0000,  0x0000 };
  GdkColor TIP_BACKGROUND_GDK = { 0, 0xffff, 0xffff,  0xb1df };

  const unsigned int STATE_MINIMIZED = 0x01;
  const unsigned int STATE_FOCUS_SET = 0x02;
  const unsigned int STATE_DOCUMENT_SET = 0x04;
  const unsigned int STATE_FUNCTION_SET = 0x08;
  const unsigned int STATE_WITHDRAWN = 0x10;
  const unsigned int STATE_PARAMTIP_ENABLED = 0x20;

  // Paramtip is only drawn if state is exactly
  // STATE_FOCUS_SET | STATE_DOCUMENT_SET | STATE_FUNCTION_SET | STATE_PARAMTIP_ENABLED
  const unsigned int DRAW_STATE = 0x2e;
}

ParamTip::ParamTip(FileHolder& file_holder, C4sPreferences* preferences, Gtk::Window& parent):
  m_file_holder(file_holder),
  m_preferences(preferences),
  m_parent(parent),
  m_tip_wnd(Gtk::WINDOW_POPUP),
  // DISABLED BY DEFAULT
  m_state(0) //preferences.show_paramtips.get() ? STATE_PARAMTIP_ENABLED : 0)
{
  //m_tip_wnd.set_window_type_hint(static_cast<Gdk::WindowTypeHint>(GDK_WINDOW_TYPE_HINT_TOOLTIP));

  m_eb.modify_bg(Gtk::STATE_NORMAL, Gdk::Color(&TIP_BACKGROUND_GDK, true));
  m_eb.set_border_width(0);
  m_eb.add(m_text);

  m_tip_wnd.modify_bg(Gtk::STATE_NORMAL, Gdk::Color(&TIP_BORDER_GDK, true));
  m_tip_wnd.set_border_width(1);
  m_tip_wnd.add(m_eb);
  m_tip_wnd.show_all_children();

  m_text.set_line_wrap(false);
  m_text.set_single_line_mode(true);

  Glib::RefPtr<Gdk::Window> window = parent.get_window();
  if(!window)
    parent.signal_realize().connect(sigc::mem_fun(*this, &ParamTip::install_event_handler));
  else
    install_event_handler();

  g_signal_connect(
    G_OBJECT(m_preferences),
    "notify",
    G_CALLBACK(on_notify),
    this
  );

  parent.signal_frame_event().connect(sigc::mem_fun(*this, &ParamTip::on_frame));
  parent.signal_focus_out_event().connect(sigc::mem_fun(*this, &ParamTip::on_focus_out));
  parent.signal_focus_in_event().connect(sigc::mem_fun(*this, &ParamTip::on_focus_in));
  parent.signal_window_state_event().connect(sigc::mem_fun(*this, &ParamTip::on_window_state));

  m_file_holder.signal_document_changed().connect(sigc::mem_fun(*this, &ParamTip::on_document_changed));
  on_document_changed(file_holder.get_current_document());
}

void ParamTip::on_notify(GObject* object, GParamSpec* pspec, gpointer data)
{
  if(strcmp(pspec->name, "view-tooltips") == 0)
    static_cast<ParamTip*>(data)->on_paramtips_changed();
}

void ParamTip::install_event_handler()
{
#ifdef _WIN32
  // Install message filter to get notified about window movement on WIN32.
  // Found no other way to perform this.
  Glib::RefPtr<Gdk::Window> window = m_parent.get_window();
  window->add_filter(on_filter, this);
#endif
}

#ifdef _WIN32
GdkFilterReturn ParamTip::on_filter(GdkXEvent* xevent, GdkEvent* event, gpointer userdata)
{
  MSG* msg = static_cast<MSG*>(xevent);
  if(msg->message == WM_WINDOWPOSCHANGED)
    static_cast<ParamTip*>(userdata)->refresh_tip();
  
  return GDK_FILTER_CONTINUE;
}
#endif

bool ParamTip::on_frame(GdkEvent* event)
{
  if(event->type == GDK_CONFIGURE)
    printf("Configure\n");
  else
    printf("Other\n");

  refresh_tip();
  return false;
}

bool ParamTip::on_focus_out(GdkEventFocus* event)
{
  // Hide tooltip to not disturb other windows
  modify_state(0, STATE_FOCUS_SET);
  return false;
}

bool ParamTip::on_focus_in(GdkEventFocus* event)
{
  // Reshow current tooltip
  modify_state(STATE_FOCUS_SET, 0);
  return false;
}

bool ParamTip::on_window_state(GdkEventWindowState* event)
{
  if(event->changed_mask & GDK_WINDOW_STATE_WITHDRAWN)
    modify_state(event->new_window_state & STATE_WITHDRAWN, ~event->new_window_state & STATE_WITHDRAWN);

  if(event->changed_mask & GDK_WINDOW_STATE_ICONIFIED)
    modify_state(event->new_window_state & STATE_MINIMIZED, ~event->new_window_state & STATE_MINIMIZED);

  if(event->changed_mask & GDK_WINDOW_STATE_MAXIMIZED)
    refresh_tip();

  if(event->changed_mask & GDK_WINDOW_STATE_FULLSCREEN)
    refresh_tip();

  return false;
}

void ParamTip::on_hscrolled()
{
  refresh_tip(m_scroll_x - m_current_document->get_hscrollbar()->get_value(), 0);
  m_scroll_x = m_current_document->get_hscrollbar()->get_value();
}

void ParamTip::on_vscrolled()
{
  refresh_tip(0, m_scroll_y - m_current_document->get_vscrollbar()->get_value());
  m_scroll_y = m_current_document->get_vscrollbar()->get_value();
}

void ParamTip::on_paramtips_changed()
{
  // DISABLED
  //modify_state(m_preferences.show_paramtips.get() ? STATE_PARAMTIP_ENABLED : 0, m_preferences.show_paramtips.get() ? 0 : STATE_PARAMTIP_ENABLED);
}

bool ParamTip::modify_state(unsigned int add, unsigned int remove)
{
  unsigned int old_state = m_state;

  m_state |= add;
  m_state &= ~remove;

  if(old_state != m_state)
    refresh_tip();

  return old_state != m_state;
}

void ParamTip::refresh_tip(int offset_x, int offset_y)
{
  if(m_state != DRAW_STATE)
  {
    m_tip_wnd.hide();
  }
  else
  {
    if(m_current_document == NULL)
    {
      throw std::logic_error(
        "ParamTip::refresh_tip:\n"
	"m_current_document is NULL event though STATE_DOCUMENT_SET is set"
      );
    }

    C4sParamInfo* info = m_current_document->get_param_info();
    const C4sFunction* func = c4s_param_info_get_function(info);

    if(func == NULL)
    {
      throw std::logic_error(
        "ParamTip::refresh_tip:\n"
	"info.get_current_function() is NULL even though STATE_FUNCTION_SET is set"
      );
    }

    show_tip(func, c4s_param_info_get_parameter(info), offset_x, offset_y);
  }
}

void ParamTip::on_document_changed(DocumentHolder* doc_holder)
{
  if(m_conn_changed.connected()) m_conn_changed.disconnect();
  if(m_conn_removed.connected()) m_conn_removed.disconnect();
  if(m_conn_hscrolled.connected()) m_conn_hscrolled.disconnect();
  if(m_conn_vscrolled.connected()) m_conn_vscrolled.disconnect();
 
  if(doc_holder == NULL)
  {
    m_current_document = NULL;
    modify_state(0, STATE_DOCUMENT_SET | STATE_FUNCTION_SET);
  }
  else
  {
    m_current_document = &doc_holder->get_document();
    modify_state(STATE_DOCUMENT_SET, 0);

    C4sParamInfo* info = m_current_document->get_param_info();
    //m_conn_changed = info.signal_function_changed().connect(sigc::hide(sigc::hide(sigc::mem_fun(*this, &ParamTip::on_function_changed))));
    //m_conn_removed = info.signal_function_removed().connect(sigc::mem_fun(*this, &ParamTip::on_function_removed));
    g_signal_connect_after(
      G_OBJECT(info),
      "changed",
      G_CALLBACK(on_function_changed_static),
      this
    );

    m_conn_vscrolled = m_current_document->get_vscrollbar()->signal_value_changed().connect(sigc::mem_fun(*this, &ParamTip::on_vscrolled));
    m_conn_hscrolled = m_current_document->get_hscrollbar()->signal_value_changed().connect(sigc::mem_fun(*this, &ParamTip::on_hscrolled));

    m_scroll_x = m_current_document->get_hscrollbar()->get_value();
    m_scroll_y = m_current_document->get_vscrollbar()->get_value();

    if(c4s_param_info_get_function(info) != NULL)
      on_function_changed();
    else
      on_function_removed();
  }
}

void ParamTip::on_function_changed_static(C4sParamInfo* info, const C4sFunction* function, guint param, gpointer user_data)
{
  ParamTip* tip = static_cast<ParamTip*>(user_data);
  if(function != NULL) tip->on_function_changed();
  else tip->on_function_removed();
}

void ParamTip::on_function_changed()
{
  if(!modify_state(STATE_FUNCTION_SET, 0))
    refresh_tip();
}

void ParamTip::on_function_removed()
{
  modify_state(0, STATE_FUNCTION_SET);
}

void ParamTip::show_tip(const C4sFunction* func, unsigned int parameter, int offset_x, int offset_y)
{
  if(m_current_document == NULL)
  {
    throw std::logic_error(
      "ParamTip::show:\n"
      "No current document available"
    );
  }

  std::string text;
  std::string offset_text;

  C4sValueType type = c4s_function_get_return_type(func);

  text = c4s_value_type_to_string(static_cast<C4sValueType>(static_cast<int>(type) & ~static_cast<int>(C4S_VALUE_TYPE_REF)));
  if(static_cast<int>(type) & static_cast<int>(C4S_VALUE_TYPE_REF)) text += "&amp;";

  text += " ";
  text += c4s_function_get_name(func);
  text += "(";

  gsize n_parameter;
  const C4sFunctionParameter* param = c4s_function_get_parameters(func, &n_parameter);

  for(unsigned int x = 0; x < n_parameter; x++)
  {
    if(x > 0) text += ", ";

    if(x == parameter)
    {
      offset_text = text;
      text += "<b>";
    }

    text += c4s_value_type_to_string(static_cast<C4sValueType>(static_cast<int>(param[x].type) & ~static_cast<int>(C4S_VALUE_TYPE_REF)));
    if(static_cast<int>(param[x].type) & static_cast<int>(C4S_VALUE_TYPE_REF)) text += "&amp;";

    text += " ";
    text += param[x].name;

    if(x == parameter) text += "</b>";   
  }
  text += ")";

  int text_width, text_height, offset_width;

  m_text.set_markup(text);
  Glib::RefPtr<Pango::Layout> layout = m_text.get_layout();
  layout->get_pixel_size(text_width, text_height);

  layout = layout->copy();
  layout->set_markup(offset_text);
  layout->get_pixel_size(offset_width, text_height);
  
  Gtk::TextView* view = m_current_document->get_text_view();
  Gtk::TextIter iter = m_current_document->get_text_buffer()->get_insert()->get_iter();

  Gdk::Rectangle location;
  view->get_iter_location(iter, location);

  int win_x, win_y;
  view->buffer_to_window_coords(Gtk::TEXT_WINDOW_WIDGET, location.get_x(), location.get_y(), win_x, win_y);
  Glib::RefPtr<Gdk::Window> win = view->get_window(Gtk::TEXT_WINDOW_WIDGET);

  int x, y;
  win->get_origin(x, y);

  m_tip_wnd.resize(text_width + 2, text_height + 2);
  m_tip_wnd.move(win_x + x - offset_width + offset_x, win_y + y + location.get_height() + offset_y);
  m_tip_wnd.show();
}
