/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-customfunction.h>
#include <c4scripter/c4s-textutil.h>
#include <c4scripter/c4s-marshal.h>

#include <gtk/gtktextbuffer.h>
#include <gtk/gtktextmark.h>

#include <string.h>

typedef struct _C4sCustomFunctionPrivate C4sCustomFunctionPrivate;
struct _C4sCustomFunctionPrivate {
  C4sCustomFunctionAccess access;
  GtkTextMark* mark_begin;
  GtkTextMark* mark_end;
  GtkTextMark* mark_access_begin;
};

#define C4S_CUSTOM_FUNCTION_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_CUSTOM_FUNCTION, \
    C4sCustomFunctionPrivate \
  ))

enum {
  PROP_0,
  PROP_ACCESS
};

enum {
  ACCESS_CHANGED = 0,
  NAME_CHANGED,
  LAST_SIGNAL
};

static C4sFunctionClass* parent_class = NULL;
static guint custom_function_signals[LAST_SIGNAL];

static gboolean
c4s_custom_function_parse_access(GtkTextIter* pos,
                                 const C4sHighlighter* highlighter,
                                 C4sCustomFunctionAccess* access)
{
  GtkTextIter end;
  gchar* acc_str;

  c4s_highlighter_backward_to_token(highlighter, pos);
  end = *pos;

  c4s_backward_identifier(pos);
  if(gtk_text_iter_compare(pos, &end) == 0)
  {
    *access = C4S_CUSTOM_FUNCTION_ACCESS_PUBLIC;
    return TRUE;
  }

  acc_str = gtk_text_iter_get_slice(pos, &end);
  if(strcmp(acc_str, "private") == 0)
    *access = C4S_CUSTOM_FUNCTION_ACCESS_PRIVATE;
  else if(strcmp(acc_str, "protected") == 0)
    *access = C4S_CUSTOM_FUNCTION_ACCESS_PROTECTED;
  else if(strcmp(acc_str, "global") == 0)
    *access = C4S_CUSTOM_FUNCTION_ACCESS_GLOBAL;
  else
    *access = C4S_CUSTOM_FUNCTION_ACCESS_PUBLIC;

  g_free(acc_str);
  return TRUE;
}

static gboolean
c4s_custom_function_parse_name(GtkTextIter* pos,
                               const C4sHighlighter* highlighter,
                               gchar** name)
{
  GtkTextIter begin;
  gchar* func_key;

  begin = *pos;
  gtk_text_iter_forward_word_end(pos);

  func_key = gtk_text_iter_get_slice(&begin, pos);
  if(strcmp(func_key, "func") != 0)
  {
    g_free(func_key);
    return FALSE;
  }

  g_free(func_key);

  if(!gtk_text_iter_starts_word(&begin)) return FALSE;
  if(!c4s_highlighter_is_code(highlighter, &begin)) return FALSE;

  /* TODO: Check for reference return */

  c4s_highlighter_forward_to_token(highlighter, pos);
  if(!g_unichar_isalpha(gtk_text_iter_get_char(pos))) return FALSE;

  begin = *pos;
  c4s_forward_identifier(pos);
  *name = gtk_text_iter_get_slice(&begin, pos);
  return TRUE;
}

static gboolean
c4s_custom_function_parse_parameters(GtkTextIter* pos,
                                     const C4sHighlighter* highlighter,
                                     C4sCustomFunction* function)
{
  gchar* first;
  gchar* second;
  GtkTextIter begin;
  C4sValueType ref;
  C4sValueType type;

  c4s_highlighter_forward_to_token(highlighter, pos);
  if(gtk_text_iter_get_char(pos) != '(') return FALSE;

  gtk_text_iter_forward_char(pos);
  c4s_highlighter_forward_to_token(highlighter, pos);

  if(gtk_text_iter_get_char(pos) == ')')
  {
    gtk_text_iter_forward_char(pos);
    return TRUE;
  }

  for(;;)
  {
    ref = 0;
    first = NULL;
    second = NULL;

    if(g_unichar_isalpha(gtk_text_iter_get_char(pos)))
    {
      begin = *pos;
      c4s_forward_identifier(pos);
      first = gtk_text_iter_get_slice(&begin, pos);

      c4s_highlighter_forward_to_token(highlighter, pos);
    }

    if(gtk_text_iter_get_char(pos) == '&')
    {
      ref = C4S_VALUE_TYPE_REF;
      gtk_text_iter_forward_char(pos);
      c4s_highlighter_forward_to_token(highlighter, pos);
    }

    if(g_unichar_isalpha(gtk_text_iter_get_char(pos)))
    {
      begin = *pos;
      c4s_forward_identifier(pos);
      second = gtk_text_iter_get_slice(&begin, pos);
      c4s_highlighter_forward_to_token(highlighter, pos);
    }

    if(first != NULL && second != NULL)
    {
      /* func foo(int bar)
       * func foo(int& bar) */
      type = c4s_value_type_from_string(first, strlen(first));

      if(type == C4S_VALUE_TYPE_UNKNOWN)
      {
        /* func foo(rofl bar) */
        g_free(first);
        g_free(second);
        return FALSE;
      }

      c4s_function_add_parameter(
        C4S_FUNCTION(function),
        type | ref,
        second
      );
    }
    else if(first != NULL && second == NULL && ref == 0)
    {
      /* func foo(bar) */
      c4s_function_add_parameter(
        C4S_FUNCTION(function),
        C4S_VALUE_TYPE_ANY | ref,
        first
      );
    }
    else if(first == NULL && second != NULL && ref == C4S_VALUE_TYPE_REF)
    {
      /* func foo(&bar) */
      c4s_function_add_parameter(
        C4S_FUNCTION(function),
        C4S_VALUE_TYPE_ANY | ref,
        second
      );
    }
    else
    {
      /* func foo(inog rofzak breb) */
      g_free(first);
      g_free(second);
      return FALSE;
    }

    if(first != NULL) g_free(first);
    if(second != NULL) g_free(second);

    if(gtk_text_iter_get_char(pos) == ')')
      break;
    else if(gtk_text_iter_get_char(pos) != ',')
      return FALSE;

    gtk_text_iter_forward_char(pos);
    c4s_highlighter_forward_to_token(highlighter, pos);

    if(gtk_text_iter_is_end(pos)) return FALSE;
  }

  /* Jump over ')' */
  gtk_text_iter_forward_char(pos);
  return TRUE;
}

static void
c4s_custom_function_access_changed_cb(C4sCustomFunction* function,
                                      C4sCustomFunctionAccess access)
{
  C4sCustomFunctionPrivate* priv;
  priv = C4S_CUSTOM_FUNCTION_PRIVATE(function);

  priv->access = access;
  g_object_notify(G_OBJECT(function), "access");
}

static void
c4s_custom_function_name_changed_cb(C4sCustomFunction* function,
                                    const gchar* name)
{
  g_object_set(G_OBJECT(function), "name", name, NULL);
}

static void
c4s_custom_function_init(C4sCustomFunction* custom_function)
{
  C4sCustomFunctionPrivate* priv;
  priv = C4S_CUSTOM_FUNCTION_PRIVATE(custom_function);

  priv->access = C4S_CUSTOM_FUNCTION_ACCESS_PUBLIC;
  priv->mark_begin = NULL;
  priv->mark_end = NULL;
  priv->mark_access_begin = NULL;
}

static void 
c4s_custom_function_finalize(GObject* object)
{
  C4sCustomFunction* custom_function;
  C4sCustomFunctionPrivate* priv;

  custom_function = C4S_CUSTOM_FUNCTION(object);
  priv = C4S_CUSTOM_FUNCTION_PRIVATE(custom_function);

  if(priv->mark_begin != NULL)
  {
    gtk_text_buffer_delete_mark(
      gtk_text_mark_get_buffer(priv->mark_begin),
      priv->mark_begin
    );
  }

  if(priv->mark_end != NULL)
  {
    gtk_text_buffer_delete_mark(
      gtk_text_mark_get_buffer(priv->mark_end),
      priv->mark_end
    );
  }

  if(priv->mark_access_begin != NULL)
  {
    gtk_text_buffer_delete_mark(
      gtk_text_mark_get_buffer(priv->mark_access_begin),
      priv->mark_access_begin
    );
  }

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_custom_function_set_property(GObject* object,
                                 guint prop_id,
                                 const GValue* value,
                                 GParamSpec* pspec)
{
  C4sCustomFunction* custom_function;
  C4sCustomFunctionPrivate* priv;

  g_return_if_fail(C4S_IS_CUSTOM_FUNCTION(object));

  custom_function = C4S_CUSTOM_FUNCTION(object);
  priv = C4S_CUSTOM_FUNCTION_PRIVATE(custom_function);

  switch(prop_id)
  {
  case PROP_ACCESS:
    priv->access = (C4sCustomFunctionAccess)g_value_get_enum(value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_custom_function_get_property(GObject* object,
                                 guint prop_id,
                                 GValue* value,
                                 GParamSpec* pspec)
{
  C4sCustomFunction* custom_function;
  C4sCustomFunctionPrivate* priv;

  g_return_if_fail(C4S_IS_CUSTOM_FUNCTION(object));

  custom_function = C4S_CUSTOM_FUNCTION(object);
  priv = C4S_CUSTOM_FUNCTION_PRIVATE(custom_function);

  switch(prop_id)
  {
  case PROP_ACCESS:
    g_value_set_enum(value, priv->access);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_custom_function_class_init(C4sCustomFunctionClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = C4S_FUNCTION_CLASS(g_type_class_peek_parent(klass));

  g_type_class_add_private(klass, sizeof(C4sCustomFunctionPrivate));

  object_class->finalize = c4s_custom_function_finalize;
  object_class->set_property = c4s_custom_function_set_property;
  object_class->get_property = c4s_custom_function_get_property;

  klass->access_changed = c4s_custom_function_access_changed_cb;
  klass->name_changed = c4s_custom_function_name_changed_cb;

  g_object_class_install_property(
    object_class,
    PROP_ACCESS,
    g_param_spec_enum(
      "access",
      "Access",
      "Function access",
      C4S_TYPE_CUSTOM_FUNCTION_ACCESS,
      C4S_CUSTOM_FUNCTION_ACCESS_PUBLIC,
      G_PARAM_READWRITE
    )
  );

  custom_function_signals[ACCESS_CHANGED] = g_signal_new(
    "access-changed",
    G_OBJECT_CLASS_TYPE(object_class),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET(C4sCustomFunctionClass, access_changed),
    NULL, NULL,
    c4s_marshal_VOID__ENUM,
    G_TYPE_NONE,
    1,
    C4S_TYPE_CUSTOM_FUNCTION_ACCESS | G_SIGNAL_TYPE_STATIC_SCOPE
  );

  custom_function_signals[NAME_CHANGED] = g_signal_new(
    "name-changed",
    G_OBJECT_CLASS_TYPE(object_class),
    G_SIGNAL_RUN_LAST,
    G_STRUCT_OFFSET(C4sCustomFunctionClass, name_changed),
    NULL, NULL,
    c4s_marshal_VOID__STRING,
    G_TYPE_NONE,
    1,
    G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE
  );
}

GType
c4s_custom_function_access_get_type(void)
{
  static GType etype = 0;
  if(etype == 0)
  {
    static const GEnumValue values[] = {
      {
        C4S_CUSTOM_FUNCTION_ACCESS_PRIVATE,
        "C4S_CUSTOM_FUNCTION_ACCESS_PRIVATE",
        "private"
      }, {
        C4S_CUSTOM_FUNCTION_ACCESS_PROTECTED,
        "C4S_CUSTOM_FUNCTION_ACCESS_PROTECTED",
        "protected"
      }, {
        C4S_CUSTOM_FUNCTION_ACCESS_PUBLIC,
        "C4S_CUSTOM_FUNCTION_ACCESS_PUBLIC",
        "public"
      }, {
        C4S_CUSTOM_FUNCTION_ACCESS_GLOBAL,
        "C4S_CUSTOM_FUNCTION_ACCESS_GLOBAL",
        "global"
      }, {
        0,
        NULL,
        NULL
      }
    };

    etype = g_enum_register_static("C4sCustomFunctionAccess", values);
  }

  return etype;
}

GType
c4s_custom_function_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sCustomFunctionClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_custom_function_class_init,
      NULL,
      NULL,
      sizeof(C4sCustomFunction),
      0,
      (GInstanceInitFunc)c4s_custom_function_init
    };

    our_type = g_type_register_static(
      C4S_TYPE_FUNCTION,
      "C4sCustomFunction",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sCustomFunctionAccess
c4s_custom_function_get_access(const C4sCustomFunction* function)
{
  return C4S_CUSTOM_FUNCTION_PRIVATE(function)->access;
}

C4sCustomFunction*
c4s_custom_function_new(GtkTextIter* pos,
                        const C4sHighlighter* highlighter)
{
  C4sCustomFunction* function;
  C4sCustomFunctionPrivate* priv;
  GtkTextBuffer* buffer;
  GtkTextIter begin;
  GtkTextIter access_begin;
  GtkTextIter own_pos;
  C4sCustomFunctionAccess access;
  gchar* name;

  begin = access_begin = own_pos = *pos;

  if(!c4s_custom_function_parse_access(&access_begin, highlighter, &access))
    return NULL;

  if(!c4s_custom_function_parse_name(&own_pos, highlighter, &name))
    return NULL;

  function = g_object_new(
    C4S_TYPE_CUSTOM_FUNCTION,
    "access", access,
    "return-type", C4S_VALUE_TYPE_ANY,
    "name", name,
    NULL
  );

  g_free(name);

  if(!c4s_custom_function_parse_parameters(&own_pos, highlighter, function))
  {
    g_object_unref(function);
    return NULL;
  }

  priv = C4S_CUSTOM_FUNCTION_PRIVATE(function);
  buffer = gtk_text_iter_get_buffer(pos);

  priv->mark_begin = gtk_text_buffer_create_mark(buffer, NULL, &begin, TRUE);
  priv->mark_end = gtk_text_buffer_create_mark(buffer, NULL, &own_pos, TRUE);

  priv->mark_access_begin = gtk_text_buffer_create_mark(
    buffer,
    NULL,
    &access_begin,
    TRUE
  );

  *pos = own_pos;
  return function;
}

gboolean
c4s_custom_function_update(C4sCustomFunction* function,
                           const C4sHighlighter* highlighter)
{
  C4sCustomFunctionPrivate* priv;
  GtkTextBuffer* buffer;
  GtkTextIter pos;
  GtkTextIter access_begin;
  C4sCustomFunctionAccess access;
  gchar* name;

  priv = C4S_CUSTOM_FUNCTION_PRIVATE(function);
  buffer = gtk_text_mark_get_buffer(priv->mark_begin);

  gtk_text_buffer_get_iter_at_mark(buffer, &pos, priv->mark_begin);
  access_begin = pos;

  if(!c4s_custom_function_parse_access(&access_begin, highlighter, &access))
    return FALSE;

  if(!c4s_custom_function_parse_name(&pos, highlighter, &name))
    return FALSE;

  c4s_function_clear_parameters(C4S_FUNCTION(function));
  if(!c4s_custom_function_parse_parameters(&pos, highlighter, function))
  {
    g_free(name);
    return FALSE;
  }

  gtk_text_buffer_move_mark(buffer, priv->mark_end, &pos);
  gtk_text_buffer_move_mark(buffer, priv->mark_access_begin, &access_begin);

  if(access != priv->access)
  {
    g_signal_emit(
      G_OBJECT(function),
      custom_function_signals[ACCESS_CHANGED],
      0,
      access
    );
  }

  if(strcmp(name, c4s_function_get_name(C4S_FUNCTION(function))) != 0)
  {
    g_signal_emit(
      G_OBJECT(function),
      custom_function_signals[NAME_CHANGED],
      0,
      name
    );
  }

  g_free(name);
  return TRUE;
}

void
c4s_custom_function_get_begin(const C4sCustomFunction* function,
                              GtkTextIter* iter)
{
  C4sCustomFunctionPrivate* priv;
  priv = C4S_CUSTOM_FUNCTION_PRIVATE(function);

  gtk_text_buffer_get_iter_at_mark(
    gtk_text_mark_get_buffer(priv->mark_begin),
    iter,
    priv->mark_begin
  );
}

void
c4s_custom_function_get_end(const C4sCustomFunction* function,
                            GtkTextIter* iter)
{
  C4sCustomFunctionPrivate* priv;
  priv = C4S_CUSTOM_FUNCTION_PRIVATE(function);

  gtk_text_buffer_get_iter_at_mark(
    gtk_text_mark_get_buffer(priv->mark_end),
    iter,
    priv->mark_end
  );
}

void
c4s_custom_function_get_access_begin(const C4sCustomFunction* function,
                                     GtkTextIter* iter)
{
  C4sCustomFunctionPrivate* priv;
  priv = C4S_CUSTOM_FUNCTION_PRIVATE(function);

  gtk_text_buffer_get_iter_at_mark(
    gtk_text_mark_get_buffer(priv->mark_access_begin),
    iter,
    priv->mark_access_begin
  );
}

gint
c4s_custom_function_compare(const C4sCustomFunction* first,
                            const C4sCustomFunction* second)
{
  GtkTextIter iter_first;
  GtkTextIter iter_second;
  
  c4s_custom_function_get_begin(first, &iter_first);
  c4s_custom_function_get_begin(second, &iter_second);
  
  return gtk_text_iter_compare(&iter_first, &iter_second);
}
