/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-function.h>

typedef struct _C4sFunctionPrivate C4sFunctionPrivate;
struct _C4sFunctionPrivate {
  C4sValueType return_type;
  gchar* name;
  C4sFunctionParameter* parameters;
  gsize n_parameters;
  gsize alloc_parameters;
};

#define C4S_FUNCTION_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_FUNCTION, \
    C4sFunctionPrivate \
  ))

enum {
  PROP_0,
  PROP_RETURN_TYPE,
  PROP_NAME
};

static GObjectClass* parent_class = NULL;

static void
c4s_function_init(C4sFunction* function)
{
  C4sFunctionPrivate* priv;
  priv = C4S_FUNCTION_PRIVATE(function);

  priv->return_type = C4S_VALUE_TYPE_ANY;
  priv->name = NULL;
  priv->parameters = g_malloc(10 * sizeof(C4sFunctionParameter));
  priv->n_parameters = 0;
  priv->alloc_parameters = 10;
}

static void 
c4s_function_finalize(GObject* object)
{
  C4sFunction* function;
  C4sFunctionPrivate* priv;
  gsize n;

  function = C4S_FUNCTION(object);
  priv = C4S_FUNCTION_PRIVATE(function);

  if(priv->name != NULL)
    g_free(priv->name);

  for(n = 0; n < priv->n_parameters; ++ n)
    g_free(priv->parameters[n].name);
  g_free(priv->parameters);

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_function_set_property(GObject* object,
                          guint prop_id,
                          const GValue* value,
                          GParamSpec* pspec)
{
  C4sFunction* function;
  C4sFunctionPrivate* priv;

  g_return_if_fail(C4S_IS_FUNCTION(object));

  function = C4S_FUNCTION(object);
  priv = C4S_FUNCTION_PRIVATE(function);

  switch(prop_id)
  {
  case PROP_RETURN_TYPE:
    priv->return_type = (C4sValueType)g_value_get_flags(value);
    break;
  case PROP_NAME:
    if(priv->name != NULL) g_free(priv->name);
    priv->name = g_value_dup_string(value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_function_get_property(GObject* object,
                          guint prop_id,
                          GValue* value,
                          GParamSpec* pspec)
{
  C4sFunction* function;
  C4sFunctionPrivate* priv;

  g_return_if_fail(C4S_IS_FUNCTION(object));

  function = C4S_FUNCTION(object);
  priv = C4S_FUNCTION_PRIVATE(function);

  switch(prop_id)
  {
  case PROP_RETURN_TYPE:
    g_value_set_flags(value, priv->return_type);
    break;
  case PROP_NAME:
    g_value_set_string(value, priv->name);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_function_class_init(C4sFunctionClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = g_type_class_peek_parent(klass);

  g_type_class_add_private(klass, sizeof(C4sFunctionPrivate));

  object_class->finalize = c4s_function_finalize;
  object_class->set_property = c4s_function_set_property;
  object_class->get_property = c4s_function_get_property;

  g_object_class_install_property(
    object_class,
    PROP_RETURN_TYPE,
    g_param_spec_flags(
      "return-type",
      "Return type",
      "Type the function returns",
      C4S_TYPE_VALUE_TYPE,
      C4S_VALUE_TYPE_ANY,
      G_PARAM_READWRITE
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_NAME,
    g_param_spec_string(
      "name",
      "Name",
      "Function name",
      NULL,
      G_PARAM_READWRITE
    )
  );
}

GType
c4s_function_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sFunctionClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_function_class_init,
      NULL,
      NULL,
      sizeof(C4sFunction),
      0,
      (GInstanceInitFunc)c4s_function_init
    };

    our_type = g_type_register_static(
      G_TYPE_OBJECT,
      "C4sFunction",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sValueType
c4s_function_get_return_type(const C4sFunction* function)
{
  return C4S_FUNCTION_PRIVATE(function)->return_type;
}

const C4sFunctionParameter*
c4s_function_get_parameters(const C4sFunction* function,
                            gsize* n_parameters)
{
  C4sFunctionPrivate* priv;
  priv = C4S_FUNCTION_PRIVATE(function);

  if(n_parameters != NULL) *n_parameters = priv->n_parameters;
  return priv->parameters;
}

const gchar*
c4s_function_get_name(const C4sFunction* function)
{
  return C4S_FUNCTION_PRIVATE(function)->name;
}

void
c4s_function_add_parameter(C4sFunction* function,
                           C4sValueType type,
                           const gchar* name)
{
  C4sFunctionPrivate* priv;
  priv = C4S_FUNCTION_PRIVATE(function);

  /* more than ten parameters, whoo */
  if(priv->n_parameters == priv->alloc_parameters)
  {
    priv->alloc_parameters += 10;
    priv->parameters = g_realloc(
      priv->parameters,
      priv->alloc_parameters * sizeof(C4sFunctionParameter)
    );
  }

  priv->parameters[priv->n_parameters].type = type;
  priv->parameters[priv->n_parameters].name = g_strdup(name);
  ++ priv->n_parameters;
}

void
c4s_function_clear_parameters(C4sFunction* function)
{
  C4sFunctionPrivate* priv;
  gsize n;

  priv = C4S_FUNCTION_PRIVATE(function);
  for(n = 0; n < priv->n_parameters; ++ n)
    g_free(priv->parameters[n].name);

  priv->n_parameters = 0;
}
