/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-http.h>

#include <zlib.h>

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

enum {
  PROP_0,

  /* construct only */
  PROP_CONNECTION,
  PROP_REQUEST_FILE,
  PROP_CB
};

typedef enum _C4sHttpRequestState {
  C4S_HTTP_REQUEST_SEND_REQUEST,
  C4S_HTTP_REQUEST_GET_HEADER,
  C4S_HTTP_REQUEST_GET_BODY,
  C4S_HTTP_REQUEST_ERROR,
  C4S_HTTP_REQUEST_DONE
} C4sHttpRequestState;

typedef enum _C4sHttpTransferEncoding {
  C4S_HTTP_REQUEST_TRANSFER_ENCODING_PLAIN,
  C4S_HTTP_REQUEST_TRANSFER_ENCODING_CHUNKED
} C4sHttpTransferEncoding;

typedef enum _C4sHttpContentEncoding {
  C4S_HTTP_REQUEST_CONTENT_ENCODING_PLAIN,
  C4S_HTTP_REQUEST_CONTENT_ENCODING_GZIPPED
} C4sHttpContentEncoding;

typedef struct _C4sHttpRequestCb C4sHttpRequestCb;
struct _C4sHttpRequestCb {
  C4sHttpDoneFunc done_func;
  C4sHttpErrorFunc error_func;
  gpointer user_data;
};

typedef struct _C4sHttpRequestPrivate C4sHttpRequestPrivate;
struct _C4sHttpRequestPrivate {
  C4sHttpRequestCb cb;
  C4sTcpConnection* conn;

#ifdef _WIN32
  guint timeout_handle;
#endif

  C4sHttpRequestState state;

  gchar* request;
  gchar* request_file;
  gsize request_pos;
  gsize request_size;

  GString* reply_header;
  gsize sep_pos;

  gsize content_length;
  C4sHttpTransferEncoding transfer_encoding;
  C4sHttpContentEncoding content_encoding;

  GString* reply_body;
  gchar chunk_buf[32 + 1];
  gsize chunk_buf_pos;

  gsize chunk_size;
  gsize chunk_pos;
};

#define C4S_HTTP_REQUEST_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_HTTP_REQUEST, \
    C4sHttpRequestPrivate \
  ))

static GObjectClass* parent_class;

static gchar*
c4s_http_request_forward_to_line_end(gchar* str)
{
  while(*str != '\n' && *str != '\r')
    ++ str;

  return str;
}

static gchar*
c4s_http_request_forward_line(gchar* str)
{
  switch(*str)
  {
  case '\r':
    ++ str;
    if(*str == '\n') ++ str;
    break;
  case '\n':
    ++ str;
    break;
  }

  return str;
}

static void
c4s_http_request_error(C4sHttpRequest* http_request,
                       const gchar* error_message,
                       ...)
{
  C4sHttpRequestPrivate* priv;
  va_list arglist;
  gchar* text;

  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);

  c4s_tcp_connection_watch(priv->conn, 0, NULL, NULL);

#ifdef _WIN32
  if(priv->timeout_handle != 0)
  {
    g_source_remove(priv->timeout_handle);
    priv->timeout_handle = 0;
  }
#endif

  va_start(arglist, error_message);
  text = g_strdup_vprintf(error_message, arglist);
  va_end(arglist);

  priv->state = C4S_HTTP_REQUEST_ERROR;
  priv->cb.error_func(http_request, text, priv->cb.user_data);

  g_free(text);
}

static void
c4s_http_request_parse_header(C4sHttpRequest* http_request)
{
  C4sHttpRequestPrivate* priv;
  GString* reply_header;
  gchar* line_begin;
  gchar* line_end;
  gchar* sep_pos;
  gsize len;

  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);
  reply_header = priv->reply_header;

  if(strncmp(reply_header->str, "HTTP/1.1 ", 9) != 0)
  {
    c4s_http_request_error(http_request, "Got invalid HTTP response");
  }
  else
  {
    line_begin = reply_header->str;
    line_end = c4s_http_request_forward_to_line_end(line_begin);

    if(strncmp(reply_header->str + 9, "200 OK", 6) != 0)
    {
      c4s_http_request_error(
        http_request,
        "Got HTTP error code %.*s",
        line_end - reply_header->str - 9,
        reply_header->str + 9
      );
    }
    else
    {
      for(;;)
      {
        /* Note that it is guaranteed that the input line is terminated
         * with a newline sequence */
        line_begin = c4s_http_request_forward_line(line_end);
        if((gsize)(line_begin - reply_header->str) == reply_header->len)
          break;

        line_end = c4s_http_request_forward_to_line_end(line_begin);
        sep_pos = memchr(line_begin, ':', line_end - line_begin);

        if(sep_pos != NULL)
        {
          len = sep_pos - line_begin;
          do { ++ sep_pos; } while(isspace(*sep_pos));

          if(strncmp(line_begin, "Content-Length", len) == 0)
          {
            priv->content_length = strtoul(
              sep_pos,
              NULL,
              10
            );
          }
          else if(strncmp(line_begin, "Transfer-Encoding", len) == 0)
          {
            if(strncmp(sep_pos, "chunked", 7) == 0)
            {
              priv->transfer_encoding = 
                C4S_HTTP_REQUEST_TRANSFER_ENCODING_CHUNKED;
            }
          }
          else if(strncmp(line_begin, "Content-Encoding", len) == 0)
          {
            if(strncmp(sep_pos, "deflate", 7) == 0 ||
               strncmp(sep_pos, "gzip", 4) == 0)
            {
              priv->content_encoding =
                C4S_HTTP_REQUEST_CONTENT_ENCODING_GZIPPED;
            }
          }
        }
      }

      priv->state = C4S_HTTP_REQUEST_GET_BODY;
    }
  }
}

static void
c4s_http_request_parse_body(C4sHttpRequest* http_request)
{
  C4sHttpRequestPrivate* priv;
  z_stream stream;
  gchar* text;
  gsize len;
  gsize alloc;
  int result;

  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);

  switch(priv->content_encoding)
  {
  case C4S_HTTP_REQUEST_CONTENT_ENCODING_PLAIN:
    /* This allows the callback handler to free the HTTP request and still
     * be able to access the reply body */
    text = priv->reply_body->str;
    len = priv->reply_body->len;

    /* new string required for finalize */
    g_string_free(priv->reply_body, FALSE);
    priv->reply_body = g_string_sized_new(0);

    priv->cb.done_func(http_request, text, len, priv->cb.user_data);

    g_free(text);
    break;
  case C4S_HTTP_REQUEST_CONTENT_ENCODING_GZIPPED:
    /* We require at least as much bytes as the encrypted text, most likely
     * some more ;) */
    alloc = priv->reply_body->len * 5;
    text = g_malloc(alloc * sizeof(gchar));
    len = 0;

    stream.next_in = (Bytef*)priv->reply_body->str;
    stream.avail_in = priv->reply_body->len;
    stream.zalloc = Z_NULL;
    stream.zfree = Z_NULL;
    stream.opaque = NULL;

    stream.next_out = (Bytef*)text;
    stream.avail_out = alloc;

    if(inflateInit2(&stream, 15 + 32) != Z_OK)
    {
      g_free(text);
      c4s_http_request_error(
        http_request,
        "Failed to initialize decompression of data sent by server"
      );
    }
    else
    {
      do
      {
        result = inflate(&stream, Z_SYNC_FLUSH);
        if(result != Z_OK && result != Z_STREAM_END)
        {
          inflateEnd(&stream);
          g_free(text);

          c4s_http_request_error(
            http_request,
            "Failed to decompress data sent by server: %s",
            stream.msg
          );

          return;
        }

        len += (alloc - len) - stream.avail_out;
        if(stream.avail_out == 0)
        {
          alloc += priv->reply_body->len;

          text = g_realloc(text, alloc * sizeof(gchar));

          stream.next_out = (Bytef*)text + len;
          stream.avail_out = alloc - len;
        }
      } while(result != Z_STREAM_END);

      inflateEnd(&stream);
      priv->cb.done_func(http_request, text, len, priv->cb.user_data);

      g_free(text);
    }

    break;
  default:
    g_assert_not_reached();
    break;
  }
}

static void
c4s_http_request_input_body(C4sHttpRequest* http_request,
                            const gchar* buffer,
                            gsize len)
{
  C4sHttpRequestPrivate* priv;
  const gchar* sep;
  const gchar* sep_pos;
  gsize copy_len;

  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);

  /* no input */
  if(len == 0); /* TODO: Return here? */

  switch(priv->transfer_encoding)
  {
  case C4S_HTTP_REQUEST_TRANSFER_ENCODING_PLAIN:
    g_string_append_len(priv->reply_body, buffer, len);

    if(priv->reply_body->len == priv->content_length)
    {
      priv->state = C4S_HTTP_REQUEST_DONE;
      c4s_tcp_connection_watch(priv->conn, 0, NULL, NULL);
    }

    break;
  case C4S_HTTP_REQUEST_TRANSFER_ENCODING_CHUNKED:
    if(priv->chunk_size == 0)
    {
      /* Chunk size not known, no read it */
      sep = memchr(buffer, '\n', len);
      sep_pos = (sep == NULL ? buffer + len : sep);

      if(priv->chunk_buf_pos + (sep_pos - buffer) > 32)
      {
        c4s_http_request_error(http_request, "Expected next chunk size");
      }
      else
      {
        strncpy(
          priv->chunk_buf + priv->chunk_buf_pos,
          buffer,
          sep_pos - buffer
        );

        priv->chunk_buf_pos += (sep_pos - buffer);
        if(sep != NULL)
        {
          priv->chunk_buf[priv->chunk_buf_pos] = '\0';

          priv->chunk_size = strtoul(priv->chunk_buf, NULL, 16);

          if(priv->chunk_size == 0)
          {
            /* TODO: Wait for terminating \r\n? */
            priv->state = C4S_HTTP_REQUEST_DONE;
            c4s_tcp_connection_watch(priv->conn, 0, NULL, NULL);
          }
          else
          {
            c4s_http_request_input_body(
              http_request,
              sep + 1,
              len - ((gsize)(sep - buffer)) - 1
            );
          }
        }
      }
    }
    else
    {
      /* chunk size known */
      if(priv->chunk_pos < priv->chunk_size)
      {
        /* We did not get the whole chunk data so far */
        copy_len = priv->chunk_size - priv->chunk_pos;

        if(copy_len > len) copy_len = len;

        g_string_append_len(priv->reply_body, buffer, copy_len);

        priv->chunk_pos += copy_len;
        if(priv->chunk_pos == priv->chunk_size)
        {
          c4s_http_request_input_body(
            http_request,
            buffer + copy_len,
            len - copy_len
          );
        }
      }
      else
      {
        /* We already got the whole chunk, but we still need to get the
         * terminating \r\n */
        copy_len = priv->chunk_size + 2 - priv->chunk_pos;
        if(copy_len > len) copy_len = len;

        priv->chunk_pos += copy_len;
        if(priv->chunk_size + 2 == priv->chunk_pos)
        {
          /* TODO: Set DONE state when chunk_size was already zero? */
          /* Reset fields for next chunk */
          priv->chunk_size = 0;
          priv->chunk_pos = 0;
          priv->chunk_buf_pos = 0;

          c4s_http_request_input_body(
            http_request,
            buffer + copy_len,
            len - copy_len
          );
        }
      }
    }

    break;
  default:
    g_assert_not_reached();
    break;
  }
}

static void
c4s_http_request_input(C4sHttpRequest* http_request,
                       const gchar* buffer,
                       gsize len)
{
  /* Separation from header and body */
  static const gchar sep[4] = "\r\n\r\n";
  static const gsize sep_len = sizeof(sep) / sizeof(sep[0]);

  C4sHttpRequestPrivate* priv;
  gsize pos;

  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);

  switch(priv->state)
  {
  case C4S_HTTP_REQUEST_SEND_REQUEST:
    /* No input expected. Normally, we should have a g_assert_not_reached()
     * assertion or something here, but the timeout handler may cause this
     * function to be called when we are in either of these states. */
    break;
  case C4S_HTTP_REQUEST_GET_HEADER:
    for(pos = 0; pos < len && priv->sep_pos < sep_len; ++ pos)
    {
      if(buffer[pos] == sep[priv->sep_pos])
        ++ priv->sep_pos;
      else if(buffer[pos] == sep[0])
        priv->sep_pos = 1;
      else
        priv->sep_pos = 0;
    }

    g_string_append_len(priv->reply_header, buffer, pos);
    if(priv->sep_pos == sep_len)
    {
      c4s_http_request_parse_header(http_request);
      if(priv->state == C4S_HTTP_REQUEST_GET_BODY)
      {
        c4s_http_request_input(http_request, buffer + pos, len - pos);
      }
    }

    break;
  case C4S_HTTP_REQUEST_GET_BODY:
    c4s_http_request_input_body(http_request, buffer, len);

    /* got whole body */
    if(priv->state == C4S_HTTP_REQUEST_DONE)
    {
      c4s_tcp_connection_watch(
        priv->conn,
        0,
        NULL,
        NULL
      );

#ifdef _WIN32
      if(priv->timeout_handle != 0)
      {
        g_source_remove(priv->timeout_handle);
        priv->timeout_handle = 0;
      }
#endif

      c4s_http_request_parse_body(http_request);
    }

    break;
  default:
    g_assert_not_reached();
    break;
  }
}

/* Forward declaration, required by c4s_http_request_output */
static void
c4s_http_request_io_cb(C4sTcpConnection* conn,
                       GIOCondition condition,
                       gpointer user_data);

static void
c4s_http_request_output(C4sHttpRequest* http_request,
                        gsize len)
{
  C4sHttpRequestPrivate* priv;
  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);

  priv->request_pos += len;
  if(priv->request_pos == priv->request_size)
  {    
    priv->state = C4S_HTTP_REQUEST_GET_HEADER;

    c4s_tcp_connection_watch(
      priv->conn,
      G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
      c4s_http_request_io_cb,
      http_request
    );
  }
}

static void
c4s_http_request_in_event(C4sHttpRequest* http_request)
{
  C4sHttpRequestPrivate* priv;
  gsize bytes_read;
  GError* error;
  gchar buffer[2048];
  GIOStatus status;

  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);
  error = NULL;

  status = c4s_tcp_connection_recv(
    priv->conn,
    buffer,
    2048,
    &bytes_read,
    &error
  );

  g_object_ref(G_OBJECT(http_request));

  /* Receive until we either get G_IO_STATUS_AGAIN or all data has been 
   * received. */
  switch(status)
  {
  case G_IO_STATUS_NORMAL:
    g_assert(bytes_read > 0);

    c4s_http_request_input(http_request, buffer, bytes_read);

    if(priv->state == C4S_HTTP_REQUEST_GET_HEADER ||
       priv->state == C4S_HTTP_REQUEST_GET_BODY)
    {
      c4s_http_request_in_event(http_request);
    }

    break;
  case G_IO_STATUS_AGAIN:
    break;
  case G_IO_STATUS_EOF:
    /* TODO: We need error codes here */
    c4s_http_request_error(http_request, "Remote site closed connection");
    break;
  case G_IO_STATUS_ERROR:
    c4s_http_request_error(http_request, "%s", error->message);
    g_error_free(error);
    break;
  default:
    g_assert_not_reached();
    break;
  }

  g_object_unref(G_OBJECT(http_request));
}

static void
c4s_http_request_out_event(C4sHttpRequest* http_request)
{
  C4sHttpRequestPrivate* priv;
  gsize bytes_written;
  GError* error;
  GIOStatus status;

  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);
  error = NULL;

  status = c4s_tcp_connection_send(
    priv->conn,
    priv->request + priv->request_pos,
    priv->request_size - priv->request_pos,
    &bytes_written,
    &error
  );

  g_object_ref(G_OBJECT(http_request));

  /* Write until we either get G_IO_STATUS_AGAIN or all data has been
   * written. */
  switch(status)
  {
  case G_IO_STATUS_NORMAL:
    g_assert(bytes_written > 0);

    c4s_http_request_output(http_request, bytes_written);

    if(priv->state == C4S_HTTP_REQUEST_SEND_REQUEST)
      c4s_http_request_out_event(http_request);

    break;
  case G_IO_STATUS_AGAIN:
    break;
  case G_IO_STATUS_EOF:
    c4s_http_request_error(http_request, "Remote site closed connection");
    break;
  case G_IO_STATUS_ERROR:
    c4s_http_request_error(http_request, "%s", error->message);
    g_error_free(error);
    break;
  default:
    g_assert_not_reached();
    break;
  }

  g_object_unref(G_OBJECT(http_request));
}

static void
c4s_http_request_io_cb(C4sTcpConnection* conn,
                       GIOCondition condition,
                       gpointer user_data)
{
  C4sHttpRequest* http_request;
  http_request = C4S_HTTP_REQUEST(user_data);

  if(condition & (G_IO_HUP | G_IO_ERR | G_IO_NVAL))
  {
    c4s_http_request_error(http_request, "Connection closed unexpectedly");
  }
  else if(condition & (G_IO_OUT))
  {
    c4s_http_request_out_event(http_request);
  }
  else if(condition & (G_IO_IN))
  {
    c4s_http_request_in_event(http_request);
  }
}

#ifdef _WIN32
static gboolean
c4s_http_request_timeout_cb(gpointer user_data)
{
  /* Moving a window while the function list is being downloaded prevents
   * glib from emitting an IO event with G_IO_IN. I do not know why, but
   * when we receive one further byte from the socket, this reenables correct
   * signal emission. So we receive bytes in the timeout handler here to
   * prevent that the whole thing hangs up when one moves the window. */
  c4s_http_request_in_event(C4S_HTTP_REQUEST(user_data));
  return TRUE;
}
#endif

static void
c4s_http_request_init(C4sHttpRequest* http_request)
{
  C4sHttpRequestPrivate* priv;
  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);

  priv->cb.done_func = NULL;
  priv->cb.error_func = NULL;
  priv->conn = NULL;
  priv->state = C4S_HTTP_REQUEST_SEND_REQUEST;
  priv->request = NULL;
  priv->request_file = NULL;
  priv->request_pos = 0;
  priv->request_size = 0;

  priv->reply_header = g_string_sized_new(1024);
  priv->sep_pos = 0;

  priv->content_length = 0;
  priv->transfer_encoding = C4S_HTTP_REQUEST_TRANSFER_ENCODING_PLAIN;
  priv->content_encoding = C4S_HTTP_REQUEST_CONTENT_ENCODING_PLAIN;

  priv->reply_body = g_string_sized_new(1024);
  priv->chunk_size = 0;
  priv->chunk_pos = 0;
}

static GObject*
c4s_http_request_constructor(GType type,
                             guint n_construct_properties,
                             GObjectConstructParam *construct_param)
{
  GObject* g_object;
  C4sHttpRequest* http_request;
  C4sHttpRequestPrivate* priv;

  g_object = G_OBJECT_CLASS(parent_class)->constructor(
    type,
    n_construct_properties,
    construct_param
  );

  http_request = C4S_HTTP_REQUEST(g_object);
  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);

  /* Must have been set as construct only property */
  g_assert(priv->conn != NULL);
  g_assert(priv->request_file != NULL);
  g_assert(priv->reply_header != NULL);

  priv->request = g_strdup_printf(
    "GET %s HTTP/1.1\r\n"
    "Host: clonk.de\r\n"
    "User-Agent: C4Scripter\r\n"
    "Accept-Encoding: gzip,deflate\r\n"
    "Connection: keep-alive\r\n\r\n",
    priv->request_file
  );

  /* TODO: Calling strlen sucks... */
  priv->request_size = strlen(priv->request);

  c4s_tcp_connection_watch(
    priv->conn,
    G_IO_OUT | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
    c4s_http_request_io_cb,
    http_request
  );

#ifdef _WIN32
  /* See explanation above in c4s_htte_request_timeout_cb */
  priv->timeout_handle = g_timeout_add(
    700,
    c4s_http_request_timeout_cb,
    http_request
  );
#endif

  return g_object;
}

static void 
c4s_http_request_finalize(GObject* object)
{
  C4sHttpRequest* http_request;
  C4sHttpRequestPrivate* priv;

  http_request = C4S_HTTP_REQUEST(object);
  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);

  if(priv->state != C4S_HTTP_REQUEST_DONE &&
     priv->state != C4S_HTTP_REQUEST_ERROR)
  {
    c4s_tcp_connection_watch(priv->conn, 0, NULL, NULL);
  }

#ifdef _WIN32
  if(priv->timeout_handle != 0)
    g_source_remove(priv->timeout_handle);
#endif

  g_string_free(priv->reply_header, TRUE);
  g_string_free(priv->reply_body, TRUE);

  g_object_unref(G_OBJECT(priv->conn));
  g_free(priv->request_file);
  g_free(priv->request);

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_http_request_set_property(GObject* object,
                              guint prop_id,
                              const GValue* value,
                              GParamSpec* pspec)
{
  C4sHttpRequest* http_request;
  C4sHttpRequestPrivate* priv;

  g_return_if_fail(C4S_IS_HTTP_REQUEST(object));

  http_request = C4S_HTTP_REQUEST(object);
  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);

  switch(prop_id)
  {
  case PROP_CONNECTION:
    if(priv->conn != NULL)
      g_object_unref(G_OBJECT(priv->conn));

    priv->conn = C4S_TCP_CONNECTION(g_value_dup_object(value));
    break;
  case PROP_REQUEST_FILE:
    priv->request_file = g_value_dup_string(value);
    break;
  case PROP_CB:
    priv->cb = *(C4sHttpRequestCb*)g_value_get_pointer(value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_http_request_get_property(GObject* object,
                                guint prop_id,
                                GValue* value,
                                GParamSpec* pspec)
{
  C4sHttpRequest* http_request;
  C4sHttpRequestPrivate* priv;

  g_return_if_fail(C4S_IS_HTTP_REQUEST(object));

  http_request = C4S_HTTP_REQUEST(object);
  priv = C4S_HTTP_REQUEST_PRIVATE(http_request);

  switch(prop_id)
  {
  case PROP_CONNECTION:
    /* TODO: Are we supposed to ref connection here? */
    g_value_set_object(value, priv->conn);
    break;
  case PROP_REQUEST_FILE:
    /* write-only */
    g_assert_not_reached();
    break;
  case PROP_CB:
    g_value_set_pointer(value, &priv->cb);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_http_request_class_init(C4sHttpRequestClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = g_type_class_peek_parent(klass);

  g_type_class_add_private(klass, sizeof(C4sHttpRequestPrivate));

  object_class->constructor = c4s_http_request_constructor;
  object_class->finalize = c4s_http_request_finalize;
  object_class->set_property = c4s_http_request_set_property;
  object_class->get_property = c4s_http_request_get_property;

  g_object_class_install_property(
    object_class,
    PROP_CONNECTION,
    g_param_spec_object(
      "connection",
      "TCP connection",
      "TCP connection to make the HTTP request with",
      C4S_TYPE_TCP_CONNECTION,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_REQUEST_FILE,
    g_param_spec_string(
      "request-file",
      "File to request",
      "File to request from the webserver",
      "",
      G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY
    )
  );

  g_object_class_install_property(
    object_class,
    PROP_CB,
    g_param_spec_pointer(
      "cb-struct",
      "Callback struct",
      "Struct defining callback functions",
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );
}

GType
c4s_http_request_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sHttpRequestClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_http_request_class_init,
      NULL,
      NULL,
      sizeof(C4sHttpRequest),
      0,
      (GInstanceInitFunc)c4s_http_request_init
    };

    our_type = g_type_register_static(
      G_TYPE_OBJECT,
      "C4sHttpRequest",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sHttpRequest*
c4s_http_request_new(C4sTcpConnection* conn,
                     const gchar* request_file,
                     C4sHttpDoneFunc done_func,
                     C4sHttpErrorFunc error_func,
                     gpointer user_data)
{
  C4sHttpRequestCb cb;
  GObject* object;

  cb.done_func = done_func;
  cb.error_func = error_func;
  cb.user_data = user_data;

  object = g_object_new(
    C4S_TYPE_HTTP_REQUEST,
    "cb-struct", &cb,
    "connection", conn,
    "request-file", request_file,
    NULL
  );

  return C4S_HTTP_REQUEST(object);
}

const gchar*
c4s_http_request_get_file(const C4sHttpRequest* request)
{
  C4sHttpRequestPrivate* priv;
  priv = C4S_HTTP_REQUEST_PRIVATE(request);
  
  return priv->request_file;
}
