/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-preferencescolors.h>

#include <gtk/gtkcolorbutton.h>
#include <gtk/gtklabel.h>
#include <gtk/gtktable.h>

#include <string.h>

typedef struct _C4sPreferencesColorsPrivate C4sPreferencesColorsPrivate;
struct _C4sPreferencesColorsPrivate
{
  C4sPreferences* preferences;
  
  GtkWidget* color_text;
  GtkWidget* color_string;
  GtkWidget* color_comment;
  GtkWidget* color_keyword;
  GtkWidget* color_function;
  GtkWidget* color_number;
};

#define C4S_PREFERENCES_COLORS_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_PREFERENCES_COLORS, \
    C4sPreferencesColorsPrivate \
  ))

enum
{
  PROP_0,

  PROP_PREFERENCES
};

static C4sPreferencesPageClass* parent_class = NULL;

static void
c4s_preferences_colors_notify_cb(GObject* object,
                                 GParamSpec* pspec,
                                 gpointer user_data);

static void
c4s_preferences_colors_color_changed_cb(GtkColorButton* button,
                                        gpointer user_data)
{
  const gchar* property_name;
  GdkColor color;

  C4sPreferencesColors* colors;
  C4sPreferencesColorsPrivate* priv;

  colors = C4S_PREFERENCES_COLORS(user_data);
  priv = C4S_PREFERENCES_COLORS_PRIVATE(colors);
  
  if(priv->preferences != NULL)
  {
    g_signal_handlers_block_by_func(
      G_OBJECT(priv->preferences),
      G_CALLBACK(c4s_preferences_colors_notify_cb),
      colors
    );

    property_name = g_object_get_data(
      G_OBJECT(button),
      "c4s-preferences-property-name"
    );
    
    gtk_color_button_get_color(button, &color);

    g_object_set(
      G_OBJECT(priv->preferences),
      property_name,
      &color,
      NULL
    );
    
    g_signal_handlers_unblock_by_func(
      G_OBJECT(priv->preferences),
      G_CALLBACK(c4s_preferences_colors_notify_cb),
      colors
    );
  }
}

static void
c4s_preferences_colors_notify_cb(G_GNUC_UNUSED GObject* object,
                                 GParamSpec* pspec,
                                 gpointer user_data)
{
  GtkWidget* button;
  GdkColor color;

  C4sPreferencesColors* colors;
  C4sPreferencesColorsPrivate* priv;

  colors = C4S_PREFERENCES_COLORS(user_data);
  priv = C4S_PREFERENCES_COLORS_PRIVATE(colors);

  if(strcmp(pspec->name, "color-text") == 0)
    button = priv->color_text;
  else if(strcmp(pspec->name, "color-string") == 0)
    button = priv->color_string;
  else if(strcmp(pspec->name, "color-comment") == 0)
    button = priv->color_comment;
  else if(strcmp(pspec->name, "color-keyword") == 0)
    button = priv->color_keyword;
  else if(strcmp(pspec->name, "color-function") == 0)
    button = priv->color_function;
  else if(strcmp(pspec->name, "color-number") == 0)
    button = priv->color_number;
  else
    return;

  g_signal_handlers_block_by_func(
    G_OBJECT(button),
    G_CALLBACK(c4s_preferences_colors_color_changed_cb),
    colors
  );

  g_object_get(
    G_OBJECT(priv->preferences),
    pspec->name,
    &color,
    NULL
  );

  gtk_color_button_set_color(GTK_COLOR_BUTTON(button), &color);

  g_signal_handlers_unblock_by_func(
    G_OBJECT(button),
    G_CALLBACK(c4s_preferences_colors_color_changed_cb),
    colors
  );
}

static GtkWidget*
c4s_preferences_colors_make_label(GtkWidget* table,
                                  guint row,
                                  const gchar* text)
{
  GtkWidget* label;
  label = gtk_label_new(text);
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
  
  gtk_table_attach(
    GTK_TABLE(table),
    label,
    0, 1, row, row + 1,
    GTK_SHRINK | GTK_FILL,
    GTK_SHRINK,
    0, 0
  );
  
  gtk_widget_show(label);
  return label;
}

static GtkWidget*
c4s_preferences_colors_make_button(GtkWidget* table,
                                   guint row,
                                   const gchar* preferences_property,
                                   gpointer signal_user_data)
{
  GtkWidget* button;
  button = gtk_color_button_new();

  g_object_set_data(
    G_OBJECT(button),
    "c4s-preferences-property-name",
    /* we could g_strdup() it, but it is static and will never be
     * changed anyway... */
    (gpointer)preferences_property
  );
  
  g_signal_connect(
    G_OBJECT(button),
    "color-set",
    G_CALLBACK(c4s_preferences_colors_color_changed_cb),
    signal_user_data
  );

  gtk_table_attach(
    GTK_TABLE(table),
    button,
    1, 2, row, row + 1,
    GTK_EXPAND | GTK_FILL,
    GTK_SHRINK,
    0, 0
  );

  gtk_widget_show(button);
  return button;
}

static void
c4s_preferences_colors_init(C4sPreferencesColors* colors)
{
  C4sPreferencesColorsPrivate* priv;
  C4sOptionGroup* group;
  GtkWidget* table;

  priv = C4S_PREFERENCES_COLORS_PRIVATE(colors);

  priv->preferences = NULL;

  table = gtk_table_new(6, 2, FALSE);
  gtk_table_set_row_spacings(GTK_TABLE(table), 8);
  gtk_table_set_col_spacings(GTK_TABLE(table), 12);

  c4s_preferences_colors_make_label(table, 0, "Normal text");
  c4s_preferences_colors_make_label(table, 1, "Strings");
  c4s_preferences_colors_make_label(table, 2, "Comments");
  c4s_preferences_colors_make_label(table, 3, "Keywords");
  c4s_preferences_colors_make_label(table, 4, "Functions");
  c4s_preferences_colors_make_label(table, 5, "Numbers");

  priv->color_text = c4s_preferences_colors_make_button(
    table,
    0,
    "color-text",
    colors
  );

  priv->color_string = c4s_preferences_colors_make_button(
    table,
    1,
    "color-string",
    colors
  );

  priv->color_comment = c4s_preferences_colors_make_button(
    table,
    2,
    "color-comment",
    colors
  );

  priv->color_keyword = c4s_preferences_colors_make_button(
    table,
    3,
    "color-keyword",
    colors
  );

  priv->color_function = c4s_preferences_colors_make_button(
    table,
    4,
    "color-function",
    colors
  );

  priv->color_number = c4s_preferences_colors_make_button(
    table,
    5,
    "color-number",
    colors
  );

  group = c4s_preferences_page_add_option_group(
    C4S_PREFERENCES_PAGE(colors),
    "Syntax highlighting"
  );

  gtk_box_pack_start(
    GTK_BOX(group->vbox),
    table,
    FALSE,
    FALSE,
    0
  );

  gtk_widget_show(table);

  gtk_widget_set_sensitive(GTK_WIDGET(colors), FALSE);
}

static void
c4s_preferences_colors_set_preferences(C4sPreferencesColors* colors,
                                       C4sPreferences* preferences)
{
  C4sPreferencesColorsPrivate* priv;
  priv = C4S_PREFERENCES_COLORS_PRIVATE(colors);

  if(priv->preferences != NULL)
  {
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(preferences),
      G_CALLBACK(c4s_preferences_colors_notify_cb),
      colors
    );

    g_object_unref(G_OBJECT(preferences));

    if(preferences == NULL)
      gtk_widget_set_sensitive(GTK_WIDGET(colors), FALSE);
  }

  if(preferences != NULL)
  {
    g_signal_connect(
      G_OBJECT(preferences),
      "notify",
      G_CALLBACK(c4s_preferences_colors_notify_cb),
      colors
    );
    
    gtk_color_button_set_color(
      GTK_COLOR_BUTTON(priv->color_text),
      c4s_preferences_get_color_text(preferences)
    );

    gtk_color_button_set_color(
      GTK_COLOR_BUTTON(priv->color_string),
      c4s_preferences_get_color_string(preferences)
    );

    gtk_color_button_set_color(
      GTK_COLOR_BUTTON(priv->color_comment),
      c4s_preferences_get_color_comment(preferences)
    );

    gtk_color_button_set_color(
      GTK_COLOR_BUTTON(priv->color_keyword),
      c4s_preferences_get_color_keyword(preferences)
    );

    gtk_color_button_set_color(
      GTK_COLOR_BUTTON(priv->color_function),
      c4s_preferences_get_color_function(preferences)
    );

    gtk_color_button_set_color(
      GTK_COLOR_BUTTON(priv->color_number),
      c4s_preferences_get_color_number(preferences)
    );

    g_object_ref(G_OBJECT(preferences));

    if(priv->preferences == NULL)
      gtk_widget_set_sensitive(GTK_WIDGET(colors), TRUE);
  }

  priv->preferences = preferences;
}

static void
c4s_preferences_colors_finalize(GObject* object)
{
  C4sPreferencesColors* colors;
  C4sPreferencesColorsPrivate* priv;
  
  colors = C4S_PREFERENCES_COLORS(object);
  priv = C4S_PREFERENCES_COLORS_PRIVATE(colors);

  if(priv->preferences != NULL)
  {
    g_signal_handlers_disconnect_by_func(
      G_OBJECT(priv->preferences),
      G_CALLBACK(c4s_preferences_colors_notify_cb),
      colors
    );

    g_object_unref(priv->preferences);
  }

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_preferences_colors_set_property(GObject* object,
                                    guint prop_id,
                                    const GValue* value,
                                    GParamSpec* pspec)
{
  C4sPreferencesColors* colors;
  C4sPreferencesColorsPrivate* priv;

  g_return_if_fail(C4S_IS_PREFERENCES_COLORS(object));

  colors = C4S_PREFERENCES_COLORS(object);
  priv = C4S_PREFERENCES_COLORS_PRIVATE(colors);

  switch(prop_id)
  {
  case PROP_PREFERENCES:
    c4s_preferences_colors_set_preferences(
      colors,
      C4S_PREFERENCES(g_value_get_object(value))
    );

    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_preferences_colors_get_property(GObject* object,
                                    guint prop_id,
                                    GValue* value,
                                    GParamSpec* pspec)
{
  C4sPreferencesColors* colors;
  C4sPreferencesColorsPrivate* priv;

  g_return_if_fail(C4S_IS_PREFERENCES_COLORS(object));

  colors = C4S_PREFERENCES_COLORS(object);
  priv = C4S_PREFERENCES_COLORS_PRIVATE(colors);

  switch(prop_id)
  {
  case PROP_PREFERENCES:
    g_value_set_object(value, priv->preferences);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_preferences_colors_class_init(C4sPreferencesColorsClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = g_type_class_peek_parent(klass);
  
  g_type_class_add_private(klass, sizeof(C4sPreferencesColorsPrivate));

  object_class->finalize = c4s_preferences_colors_finalize;
  object_class->set_property = c4s_preferences_colors_set_property;
  object_class->get_property = c4s_preferences_colors_get_property;

  g_object_class_install_property(
    object_class,
    PROP_PREFERENCES,
    g_param_spec_object(
      "preferences",
      "Preferences",
      "Preferences to operate on",
      C4S_TYPE_PREFERENCES,
      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
    )
  );
}

GType
c4s_preferences_colors_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sPreferencesColorsClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_preferences_colors_class_init,
      NULL,
      NULL,
      sizeof(C4sPreferencesColors),
      0,
      (GInstanceInitFunc)c4s_preferences_colors_init,
      NULL
    };

    our_type = g_type_register_static(
      C4S_TYPE_PREFERENCES_PAGE,
      "C4sPreferencesColors",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sPreferencesColors*
c4s_preferences_colors_new(C4sPreferences* preferences)
{
  GObject* object;

  object = g_object_new(
    C4S_TYPE_PREFERENCES_COLORS,
    "preferences", preferences,
    NULL
  );

  return C4S_PREFERENCES_COLORS(object);
}
