/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-preferencesdialog.h>

#include <c4scripter/c4s-preferencessettings.h>
#include <c4scripter/c4s-preferencescolors.h>
#include <c4scripter/c4s-preferencesfonts.h>
#include <c4scripter/c4s-preferencesdocs.h>

#include <gtk/gtknotebook.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkstock.h>

typedef struct _C4sPreferencesDialogPrivate C4sPreferencesDialogPrivate;
struct _C4sPreferencesDialogPrivate {
  C4sPreferences* preferences;
  C4sEngineFunctionList* engine_function_list;

  C4sPreferencesSettings* settings;
  C4sPreferencesColors* colors;
  C4sPreferencesFonts* fonts;
  C4sPreferencesDocs* docs;
};

#define C4S_PREFERENCES_DIALOG_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_PREFERENCES_DIALOG, \
    C4sPreferencesDialogPrivate \
  ))

enum {
  PROP_0,

  /* construct only */
  PROP_PREFERENCES,
  PROP_ENGINE_FUNCTION_LIST
};

static GtkDialogClass* parent_class = NULL;

static void
c4s_preferences_dialog_init(C4sPreferencesDialog* dialog)
{
  C4sPreferencesDialogPrivate* priv;
  priv = C4S_PREFERENCES_DIALOG_PRIVATE(dialog);
  
  priv->preferences = NULL;
  priv->engine_function_list = NULL;

  gtk_dialog_add_button(
    GTK_DIALOG(dialog),
    GTK_STOCK_CLOSE,
    GTK_RESPONSE_CLOSE
  );
}

static GObject*
c4s_preferences_dialog_constructor(GType type,
                                   guint n_construct_properties,
                                   GObjectConstructParam *construct_param)
{
  GObject* g_object;
  C4sPreferencesDialog* dialog;
  C4sPreferencesDialogPrivate* priv;
  GtkWidget* notebook;

  g_object = G_OBJECT_CLASS(parent_class)->constructor(
    type,
    n_construct_properties,
    construct_param
  );

  dialog = C4S_PREFERENCES_DIALOG(g_object);
  priv = C4S_PREFERENCES_DIALOG_PRIVATE(dialog);

  priv->settings = c4s_preferences_settings_new(priv->preferences);
  priv->colors = c4s_preferences_colors_new(priv->preferences);
  priv->fonts = c4s_preferences_fonts_new(priv->preferences);

  priv->docs = c4s_preferences_docs_new(
    priv->preferences,
    priv->engine_function_list
  );

  notebook = gtk_notebook_new();
  
  gtk_notebook_append_page(
    GTK_NOTEBOOK(notebook),
    GTK_WIDGET(priv->settings),
    gtk_label_new("Settings")
  );

  gtk_notebook_append_page(
    GTK_NOTEBOOK(notebook),
    GTK_WIDGET(priv->colors),
    gtk_label_new("Colors")
  );

  gtk_notebook_append_page(
    GTK_NOTEBOOK(notebook),
    GTK_WIDGET(priv->fonts),
    gtk_label_new("Fonts")
  );

  gtk_notebook_append_page(
    GTK_NOTEBOOK(notebook),
    GTK_WIDGET(priv->docs),
    gtk_label_new("Documentation")
  );
  
  gtk_widget_show(GTK_WIDGET(priv->settings));
  gtk_widget_show(GTK_WIDGET(priv->colors));
  gtk_widget_show(GTK_WIDGET(priv->fonts));
  gtk_widget_show(GTK_WIDGET(priv->docs));

  gtk_container_set_border_width(GTK_CONTAINER(notebook), 6);
  gtk_widget_show(notebook);
  
  gtk_box_pack_start(
    GTK_BOX(GTK_DIALOG(dialog)->vbox),
    notebook,
    TRUE,
    TRUE,
    0
  );

  return g_object;
}


static void 
c4s_preferences_dialog_finalize(GObject* object)
{
  C4sPreferencesDialog* preferences_dialog;
  C4sPreferencesDialogPrivate* priv;

  preferences_dialog = C4S_PREFERENCES_DIALOG(object);
  priv = C4S_PREFERENCES_DIALOG_PRIVATE(preferences_dialog);

  if(priv->preferences != NULL)
    g_object_unref(G_OBJECT(priv->preferences));

  if(priv->engine_function_list != NULL)
    g_object_unref(G_OBJECT(priv->engine_function_list));

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_preferences_dialog_set_property(GObject* object,
                                       guint prop_id,
                                       const GValue* value,
                                       GParamSpec* pspec)
{
  C4sPreferencesDialog* dialog;
  C4sPreferencesDialogPrivate* priv;

  g_return_if_fail(C4S_IS_PREFERENCES_DIALOG(object));

  dialog = C4S_PREFERENCES_DIALOG(object);
  priv = C4S_PREFERENCES_DIALOG_PRIVATE(dialog);

  switch(prop_id)
  {
  case PROP_PREFERENCES:
    if(priv->preferences != NULL)
      g_object_unref(G_OBJECT(priv->preferences));

    priv->preferences = C4S_PREFERENCES(g_value_dup_object(value));
    /* TODO: We probably should update preferences of pages as well */
    break;
  case PROP_ENGINE_FUNCTION_LIST:
    if(priv->engine_function_list != NULL)
      g_object_unref(G_OBJECT(priv->engine_function_list));
      
    priv->engine_function_list = C4S_ENGINE_FUNCTION_LIST(
      g_value_dup_object(value)
    );
    
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_preferences_dialog_get_property(GObject* object,
                                    guint prop_id,
                                    G_GNUC_UNUSED GValue* value,
                                    GParamSpec* pspec)
{
  C4sPreferencesDialog* dialog;
  C4sPreferencesDialogPrivate* priv;

  g_return_if_fail(C4S_IS_PREFERENCES_DIALOG(object));

  dialog = C4S_PREFERENCES_DIALOG(object);
  priv = C4S_PREFERENCES_DIALOG_PRIVATE(dialog);

  switch(prop_id)
  {
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_preferences_dialog_response(GtkDialog* dialog,
                                int response_id)
{
  if(parent_class->response != NULL)
    parent_class->response(dialog, response_id);

  if(response_id == GTK_RESPONSE_CLOSE)
    gtk_widget_hide(GTK_WIDGET(dialog));
}

static void
c4s_preferences_dialog_class_init(C4sPreferencesDialogClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = GTK_DIALOG_CLASS(g_type_class_peek_parent(klass));

  g_type_class_add_private(klass, sizeof(C4sPreferencesDialogPrivate));

  object_class->constructor = c4s_preferences_dialog_constructor;
  object_class->finalize = c4s_preferences_dialog_finalize;
  object_class->set_property = c4s_preferences_dialog_set_property;
  object_class->get_property = c4s_preferences_dialog_get_property;
  
  GTK_DIALOG_CLASS(klass)->response = c4s_preferences_dialog_response;

  g_object_class_install_property(
    object_class,
    PROP_PREFERENCES,
    g_param_spec_object(
      "preferences",
      "Preferences",
      "Preferences to operate on",
      C4S_TYPE_PREFERENCES,
      G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY
    )
  );
  
  g_object_class_install_property(
    object_class,
    PROP_ENGINE_FUNCTION_LIST,
    g_param_spec_object(
      "engine-function-list",
      "Engine function list",
      "List of engine functions that may be reloaded",
      C4S_TYPE_ENGINE_FUNCTION_LIST,
      G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY
    )
  );
}

GType
c4s_preferences_dialog_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sPreferencesDialogClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_preferences_dialog_class_init,
      NULL,
      NULL,
      sizeof(C4sPreferencesDialog),
      0,
      (GInstanceInitFunc)c4s_preferences_dialog_init,
      NULL
    };

    our_type = g_type_register_static(
      GTK_TYPE_DIALOG,
      "C4sPreferencesDialog",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sPreferencesDialog*
c4s_preferences_dialog_new(GtkWindow* parent,
                           GtkDialogFlags flags,
                           C4sPreferences* preferences,
                           C4sEngineFunctionList* engine_function_list)
{
  GObject* object;

  object = g_object_new(
    C4S_TYPE_PREFERENCES_DIALOG,
    "preferences", preferences,
    "engine-function-list", engine_function_list,
    NULL
  );

  if(flags & GTK_DIALOG_MODAL)
    gtk_window_set_modal(GTK_WINDOW(object), TRUE);

  if(flags & GTK_DIALOG_DESTROY_WITH_PARENT)
    gtk_window_set_destroy_with_parent(GTK_WINDOW(object), TRUE);

  if(flags & GTK_DIALOG_NO_SEPARATOR)
    gtk_dialog_set_has_separator(GTK_DIALOG(object), FALSE);

  gtk_window_set_transient_for(GTK_WINDOW(object), parent);

  return C4S_PREFERENCES_DIALOG(object);
}
