/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-enginefunctionlist.h>
#include <c4scripter/c4s-preferences.h>

#include <glib/gerror.h>

#include <chm_lib.h>

#include <string.h>

struct _C4sLoaderBackend {
  struct chmFile* chm_file;
  GPtrArray* function_paths;

  C4sPreferences* preferences;
  const C4sLoaderBackendCallbacks* callbacks;
  gpointer user_data;

  unsigned int load_function;

  guint idle_handler;
};

static const gsize CHM_FUNC_PREFIX = 15; /* strlen("/sdk/script/fn/") */

static void
c4s_loader_backend_chm_set_status(C4sLoaderBackend* backend,
                                  const char* format,
                                  ...)
{
  gchar* status_str;
  va_list arglist;
  
  va_start(arglist, format);
  status_str = g_strdup_vprintf(format, arglist);
  va_end(arglist);
  
  backend->callbacks->set_status(status_str, backend->user_data);
  g_free(status_str);
}

static int
c4s_loader_backend_chm_enumerator(struct chmFile* file,
                                  struct chmUnitInfo* info,
                                  void* context)
{
  C4sLoaderBackend* backend;
  backend = context;

  /* Exclude return, for, if and while, these are keywords rather than
   * functions. */
  if(strncmp(info->path + CHM_FUNC_PREFIX, "return.", 7) == 0)
    return CHM_ENUMERATOR_CONTINUE;
  if(strncmp(info->path + CHM_FUNC_PREFIX, "while.", 6) == 0)
    return CHM_ENUMERATOR_CONTINUE;
  if(strncmp(info->path + CHM_FUNC_PREFIX, "for.", 4) == 0)
    return CHM_ENUMERATOR_CONTINUE;
  if(strncmp(info->path + CHM_FUNC_PREFIX, "if.", 3) == 0)
    return CHM_ENUMERATOR_CONTINUE;

  g_ptr_array_add(backend->function_paths, g_strdup(info->path));
  return CHM_ENUMERATOR_CONTINUE;
}

static gboolean
c4s_loader_backend_chm_init_idle(gpointer user_data)
{
  C4sLoaderBackend* backend;
  GError* error;

  backend = user_data;
  error = NULL;

  backend->idle_handler = 0;
  backend->chm_file = chm_open(
    c4s_preferences_get_chm_file_path(backend->preferences)
  );

  if(backend->chm_file == NULL)
  {
    g_set_error(
      &error,
      g_quark_from_static_string("C4S_LOADER_BACKEND_CHM"),
      C4S_LOADER_TRANSPORT_ERROR,
      "Could not open %s",
      c4s_preferences_get_chm_file_path(backend->preferences)
    );

    backend->callbacks->init_failed(error, backend->user_data);
    g_error_free(error);
  }
  else
  {
    /* TODO: Might this fail? */
    chm_enumerate_dir(
      backend->chm_file,
      "/sdk/script/fn",
      CHM_ENUMERATE_NORMAL,
      c4s_loader_backend_chm_enumerator,
      backend
    );

    backend->callbacks->init_done(
      backend->function_paths->len,
      backend->user_data
    );
  }

  return FALSE;
}

static gboolean
c4s_loader_backend_chm_load_idle(gpointer user_data)
{
  C4sLoaderBackend* backend;
  C4sEngineFunction* function;
  struct chmUnitInfo info;
  guint8* buffer;
  GError* error;
  int result;

  const gchar* path;
  const gchar* sep;
  gchar* name;

  backend = user_data;
  path = backend->function_paths->pdata[backend->load_function];
  error = NULL;
  result = chm_resolve_object(backend->chm_file, path, &info);

  sep = strrchr(path + CHM_FUNC_PREFIX, '.');
  g_assert(sep != NULL);

  name = g_strndup(path + CHM_FUNC_PREFIX, sep - path - CHM_FUNC_PREFIX);

  backend->load_function = 0;
  backend->idle_handler = 0;

  if(result == CHM_RESOLVE_FAILURE)
  {
    g_set_error(
      &error,
      g_quark_from_static_string("C4S_LOADER_BACKEND_CHM"),
      C4S_LOADER_PARSE_ERROR,
      "Could not resolve %s",
      path
    );
    
    backend->callbacks->function_failed(name, error, backend->user_data);
    g_error_free(error);
  }
  else
  {
    buffer = g_new(guint8, info.length);
    chm_retrieve_object(backend->chm_file, &info, buffer, 0, info.length);

    function = c4s_engine_function_new((const gchar*)buffer, info.length);
    if(function == NULL)
    {
      g_set_error(
        &error,
        g_quark_from_static_string("C4S_LOADER_BACKEND_CHM"),
        C4S_LOADER_PARSE_ERROR,
        "Failed to parse %s",
        path
      );
      
      backend->callbacks->function_failed(name, error, backend->user_data);
      g_error_free(error);
    }
    else
    {
      backend->callbacks->function_done(function, backend->user_data);
    }

    g_free(buffer);
  }

  g_free(name);
  return FALSE;
}

static C4sLoaderBackend*
c4s_loader_backend_chm_init(C4sPreferences* preferences,
                            const C4sLoaderBackendCallbacks* callbacks,
                            gpointer user_data)
{
  C4sLoaderBackend* backend;
  backend = g_new(C4sLoaderBackend, 1);
	
  backend->chm_file = NULL;
  backend->function_paths = g_ptr_array_new();

  backend->preferences = preferences;
  backend->callbacks = callbacks;
  backend->user_data = user_data;

  backend->load_function = 0;
  
  g_object_ref(G_OBJECT(backend->preferences));
  
  backend->idle_handler = g_idle_add(
    c4s_loader_backend_chm_init_idle,
    backend
  );

  c4s_loader_backend_chm_set_status(
    backend,
    "Opening %s",
    c4s_preferences_get_chm_file_path(preferences)
  );

  return backend;
}

static void
c4s_loader_backend_chm_deinit(C4sLoaderBackend* backend)
{
  gsize i;

  if(backend->idle_handler != 0)
    g_source_remove(backend->idle_handler);

  if(backend->chm_file != NULL)
    chm_close(backend->chm_file);

  g_object_unref(G_OBJECT(backend->preferences));
  
  for(i = 0; i < backend->function_paths->len; ++ i)
    g_free(backend->function_paths->pdata[i]);
  g_ptr_array_free(backend->function_paths, TRUE);
  g_free(backend);
}

static void
c4s_loader_backend_chm_load_function(C4sLoaderBackend* backend,
                                     unsigned int i)
{
  g_assert(backend->load_function == 0);

  backend->load_function = i;

  backend->idle_handler = g_idle_add(
    c4s_loader_backend_chm_load_idle,
    backend
  );
}

const C4sLoaderBackendFuncs C4S_LOADER_BACKEND_CHM = {
  c4s_loader_backend_chm_init,
  c4s_loader_backend_chm_deinit,
  c4s_loader_backend_chm_load_function
};
